from scipy.optimize._numdiff import approx_derivative
from scipy import interpolate
import numpy as np

from .. import math
from . import flow_model as fm
from . import deviation_model as dm

CHOKING_MODELS = [
    "evaluate_cascade_critical",
    "evaluate_cascade_throat",
    "evaluate_cascade_isentropic_throat",
]


def evaluate_choking(
    choking_input,
    inlet_plane,
    exit_plane,
    fluid,
    geometry,
    angular_speed,
    model_options,
    reference_values,
):
    r"""

    Calculate condition for choking and evaluate wheter or not the cascade is choked, based on selected choking model.

    Parameters
    ----------
    choking_input : dict
        Dictionary containing necessary input parameters required for selected choking model. See each models documentation for specifics.
    inlet_plane : dict
        Dictionary containing data on the inlet plane for the actual cascade operating condition.
    exit_plane : dict
        Dictionary containing data on the exit plane for the actual cascade operating condition.
    fluid : object
        A fluid object with methods for thermodynamic property calculations.
    geometry : dict
        Geometric parameters of the cascade.
    angular_speed : float
        Angular speed of the cascade.
    model_options : dict
        Options for the model used to evaluate choking.
    reference_values : dict
        Reference values used in the calculation, including the reference mass flow rate.

    Returns
    -------
    dict
        Dictionary containing the residuals of choking model
    dict
        Dictionary containing state information on the critical conditions.

    Raises
    ------
    ValueError
        If an invalid choking model is provided.

    """

    critical_cascade_functions = {
        CHOKING_MODELS[0]: evaluate_cascade_critical,
        CHOKING_MODELS[1]: evaluate_cascade_throat,
        CHOKING_MODELS[2]: evaluate_cascade_isentropic_throat,
    }

    # Evaluate loss model
    model = model_options["choking_model"]
    if model in critical_cascade_functions:
        return critical_cascade_functions[model](
            choking_input,
            inlet_plane,
            exit_plane,
            fluid,
            geometry,
            angular_speed,
            model_options,
            reference_values,
        )
    else:
        options = ", ".join(f"'{k}'" for k in CHOKING_MODELS)
        raise ValueError(
            f"Invalid critical cascade model '{model}'. Available options: {options}"
        )


def evaluate_cascade_throat(
    choking_input,
    inlet_plane,
    exit_plane,
    fluid,
    geometry,
    angular_speed,
    model_options,
    reference_values,
):
    r"""

    Calculate condition for choking and evaluate wheter or not the cascade is choked, based on the evaluate_cascade_throat choking model.

    This choking model evaluates the cascade throat and checks if the throat mach number exceed the critical. The critical mach number is calculated
    from a correlation depending on the throat loss coefficient. The exit flow angle is calculated by the selected deviation model at subsonic condition,
    and by ensuring that the mach at the throat equals the critical at supercritical conditions.

    Parameters
    ----------
    choking_input : dict
        Dictionary containing scaled input parameters required for selected choking model. Required items are:

        - `w*_throat` : throat velocity.
        - `s*_throat` : throat entropy.
        - `beta*_throat` : throat relative flow angle.
    inlet_plane : dict
        Dictionary containing data on the inlet plane for the actual cascade operating condition.
    exit_plane : dict
        Dictionary containing data on the exit plane for the actual cascade operating condition.
    fluid : object
        A fluid object with methods for thermodynamic property calculations.
    geometry : dict
        Geometric parameters of the cascade.
    angular_speed : float
        Angular speed of the cascade.
    model_options : dict
        Options for the model used to evaluate choking.
    reference_values : dict
        Reference values used in the calculation, including the reference mass flow rate.

    Returns
    -------
    dict
        Dictionary containing the residuals of choking model:

        - `m*`: Mass flow rate residual.
        - `Y*`: Loss error residual.
        - `beta*`: Residual of the flow angle.
        - `choking`: Choking residual.
    dict
        Dictionary containing relevant information on the critical state and throat plane.

    """

    # Rename variables
    loss_model = model_options["loss_model"]
    blockage = model_options["blockage_model"]
    deviation_model = model_options["deviation_model"]
    A_throat = geometry["A_throat"]
    A_out = geometry["A_out"]
    v0 = reference_values["v0"]
    s_range = reference_values["s_range"]
    s_min = reference_values["s_min"]
    angle_range = reference_values["angle_range"]
    angle_min = reference_values["angle_min"]

    # Evaluate throat
    cascade_throat_input = {
        "w": choking_input["w*_throat"] * v0,
        "s": choking_input["s*_throat"] * s_range + s_min,
        "beta": choking_input["beta*_throat"] * angle_range + angle_min,
        "rothalpy": inlet_plane["rothalpy"],
    }
    throat_plane, loss_dict = fm.evaluate_cascade_throat(
        cascade_throat_input,
        fluid,
        geometry,
        inlet_plane,
        angular_speed,
        blockage,
        loss_model,
    )

    # Evaluate critical mach
    Y_tot = loss_dict["loss_total"]
    critical_mass_flux, critical_mach = interpolate_critical_state(
        inlet_plane["p0_rel"], inlet_plane["T0_rel"], Y_tot
    )

    # Evaluate residual flow angle
    beta_model = np.sign(throat_plane["beta"]) * dm.get_subsonic_deviation(
        throat_plane["Ma_rel"], critical_mach, geometry, deviation_model
    )
    beta_residual = math.cosd(beta_model) - math.cosd(throat_plane["beta"])

    # Evaluate if flow cascade is choked or not an add choking residual
    if exit_plane["Ma_rel"] <= critical_mach:
        beta_model = np.sign(exit_plane["beta"]) * dm.get_subsonic_deviation(
            exit_plane["Ma_rel"], critical_mach, geometry, deviation_model
        )
        # Compute error of guessed beta and deviation model
        choking_residual = math.cosd(beta_model) - math.cosd(exit_plane["beta"])
    else:
        choking_residual = throat_plane["Ma_rel"] - critical_mach

    # Evaluate resiudals
    residual_values = np.array(
        [
            (inlet_plane["mass_flow"] - throat_plane["mass_flow"])
            / reference_values["mass_flow_ref"],
            throat_plane["loss_error"],
            beta_residual,
            choking_residual,
        ]
    )
    residual_keys = ["m*", "Y*", "beta*", "choking"]
    residuals_critical = dict(zip(residual_keys, residual_values))

    # Define output values
    critical_state = {
        "critical_mach": critical_mach,
        "critical_mass_flow_rate": critical_mass_flux * A_throat,
    }
    throat_plane = {f"{key}_throat": val for key, val in throat_plane.items()}

    return residuals_critical, {**critical_state, **throat_plane}


def interpolate_critical_state(p0_in, T0_in, Y):
    r"""
    Compute the critical mass flux and mach number from a correlation depending on the cascade inlet stagnation state and loss coefficient.
    The correlation is made from linear regression analysis, and the regression coefficients are prescribed in the function.

    Parameters
    ----------
    p0_in : float
        Cascade inlet stagnation pressure.
    T0_in : float
        Cascade inlet stagnation temperature.
    Y : float
        Cascade throat total loss coefficnet.

    Returns
    -------
    float
        Critical mass flux.
    float
        Critical mach number.

    """

    x = np.array(
        [
            1,
            p0_in,
            T0_in,
            Y,
            p0_in**2,
            T0_in**2,
            Y**2,
            p0_in * T0_in,
            p0_in * Y,
            T0_in * Y,
        ]
    )

    coeff_mach_crit = np.array(
        [
            9.97808878e-01,
            -8.59556818e-09,
            2.18283101e-05,
            -3.38413836e-01,
            -4.89469816e-14,
            -5.99021408e-08,
            9.93519991e-02,
            7.71201115e-11,
            -4.13346725e-09,
            4.91317761e-06,
        ]
    )

    coeff_phi_max = np.array(
        [
            9.81120337e01,
            3.46299580e-03,
            -6.34357717e-01,
            -6.84234362e01,
            1.05506996e-11,
            1.04045797e-03,
            3.36231786e01,
            -3.81019918e-06,
            -6.90348074e-04,
            1.26692586e-01,
        ]
    )

    mach_crit = sum(x * coeff_mach_crit)
    phi_max = sum(x * coeff_phi_max)

    return phi_max, mach_crit


def evaluate_cascade_critical(
    choking_input,
    inlet_plane,
    exit_plane,
    fluid,
    geometry,
    angular_speed,
    model_options,
    reference_values,
):
    r"""

    Calculate condition for choking and evaluate wheter or not the cascade is choked, based on the `evaluate_cascade_critical` choking model.

    This choking model evaluate the critical state by optimizing the mass flow rate at the throat through the method of lagrange multipliers, and checks if the throat mach number exceed the critical. 
    The exit flow angle is calculated by the selected devaition model at subsonic condition, and from the critical mass flow rate at supersonic conditions. 

    Compute the gradient of the Lagrange function of the critical mass flow rate and the residuals of mass
    conservation and loss computation equations at the throat.

    The constrained optimization problem to maximize the mass flow rate can be converted to a set of equations through the method of lagrange multipliers. The method
    utlize that at the optimal point, the gradient of the constraints are proportional to the gradient fo the objective function. Thus the solution to the optimization problem 
    eaulas the solution as the following set of equations: 
        
    .. math::

        \nabla L = \begin{bmatrix}
        \frac{\partial f}{\partial x_1} + \lambda_1 \cdot \frac{\partial g_1}{\partial x_1} + \lambda_2 \cdot \frac{\partial g_2}{\partial x_1} \\
        \frac{\partial f}{\partial x_2} + \lambda_1 \cdot \frac{\partial g_1}{\partial x_2} + \lambda_2 \cdot \frac{\partial g_2}{\partial x_2} \\
        \frac{\partial f}{\partial x_3} + \lambda_1 \cdot \frac{\partial g_1}{\partial x_3} + \lambda_2 \cdot \frac{\partial g_2}{\partial x_3} \\
        g_1(x_1, x_2, x_3) \\
        g_2(x_1, x_2, x_3) \\
        \end{bmatrix} = 0
        
    This function returns the value of the three last equations in the set above, while the two first equations are used to 
    explicitly calculate the lagrange multipliers.
    
    .. math::
        \lambda_1 = \frac{\frac{\partial g_2}{\partial x_2}\cdot\frac{-\partial f}{\partial x_1} - \frac{\partial g_2}{\partial x_1}\cdot-\frac{\partial f}{\partial x_2}} 
                     {\frac{\partial g_1}{\partial x_1}\cdot\frac{\partial g_2}{\partial x_2} - \frac{\partial g_2}{\partial x_1}\cdot\frac{\partial g_1}{\partial x_2}} \\                     
        \lambda_2 = \frac{\frac{\partial g_1}{\partial x_1}\cdot\frac{-\partial f}{\partial x_2} - \frac{\partial g_1}{\partial x_2}\cdot-\frac{\partial f}{\partial x_1}} 
                     {\frac{\partial g_1}{\partial x_1}\cdot\frac{\partial g_2}{\partial x_2} - \frac{\partial g_2}{\partial x_1}\cdot\frac{\partial g_1}{\partial x_2}} 

    By transforming the problem into a system of equations, this approach allows the evaluation of the critical
    point without directly solving an optimization problem. One significant advantage of this
    equation-oriented method is that it enables the coupling of these critical condition equations with the
    other modeling equations. This integrated system of equations can then be efficiently solved using gradient-based
    root finding algorithms (e.g., Newton-Raphson solvers).

    Such a coupled solution strategy, as opposed to segregated approaches where nested systems are solved
    sequentially and iteratively, offers superior computational efficiency. This method thus provides
    a more direct and computationally effective way of determining the critical conditions in a cascade.

    Parameters
    ----------
    critical_cascade_input : dict
        Dictionary containing scaled input parameters required for selected choking model. Required items are:

        - `v*_in` : inlet velocity at critical state.  
        - `w*_throat` : throat velocity at critical state. 
        - `s*_throat` : throat entropy at critical state.
    inlet_plane : dict
        Dictionary containing data on the inlet plane for the actual cascade operating condition.
    exit_plane : dict
        Dictionary containing data on the exit plane for the actual cascade operating condition.
    fluid : object
        A fluid object with methods for thermodynamic property calculations.
    geometry : dict
        Geometric parameters of the cascade.
    angular_speed : float
        Angular speed of the cascade.
    model_options : dict
        Options for the model used in the critical condition evaluation.
    reference_values : dict
        Reference values used in the calculation, including the reference mass flow rate.

    Returns
    -------
    dict
        Dictionary containing the residuals of choking model:
        
        - `m*`: Mass flow rate residual.
        - `Y*`: Loss error residual.
        - `L*`: Residual of the lagrange function.
        - `choking`: Choking residual.
    dict
        Dictionary containing state information on the critical conditions at inlet and throat plane. 

    """

    # Load reference values
    mass_flow_ref = reference_values["mass_flow_ref"]

    # Load model options
    loss_model = model_options["loss_model"]
    blockage = model_options["blockage_model"]
    deviation_model = model_options["deviation_model"]

    # Define critical state dictionary to store information
    critical_state = {}

    # Define array for inputs in compute critical values
    x_crit = np.array(
        [
            choking_input["v*_in"],
            choking_input["w*_throat"],
            choking_input["s*_throat"],
        ]
    )

    # Evaluate the current cascade at critical conditions
    f0 = compute_critical_values(
        x_crit,
        inlet_plane,
        fluid,
        geometry,
        angular_speed,
        critical_state,
        model_options,
        reference_values,
    )

    # Evaluate the Jacobian of the evaluate_critical_cascade function
    J = compute_critical_jacobian(
        x_crit,
        inlet_plane,
        fluid,
        geometry,
        angular_speed,
        critical_state,
        model_options,
        reference_values,
        f0,
    )

    # Rename gradients
    a11, a12, a21, a22, b1, b2 = (
        J[1, 0],
        J[2, 0],
        J[1, 1 + 1],
        J[2, 1 + 1],
        -1 * J[0, 0],
        -1 * J[0, 1 + 1],
    )

    # Calculate the Lagrange multipliers explicitly
    determinant = a11 * a22 - a12 * a21
    l1_det = a22 * b1 - a12 * b2
    l2_det = a11 * b2 - a21 * b1

    # Evaluate the last equation
    df, dg1, dg2 = J[0, 2 - 1], J[1, 2 - 1], J[2, 2 - 1]
    grad = (determinant * df + l1_det * dg1 + l2_det * dg2) / mass_flow_ref

    critical_mach = critical_state["throat_plane"]["Ma_rel"]
    critical_mass_flow = critical_state["throat_plane"]["mass_flow"]
    if exit_plane["Ma_rel"] <= critical_mach:
        beta_model = np.sign(exit_plane["beta"]) * dm.get_subsonic_deviation(
            exit_plane["Ma_rel"], critical_mach, geometry, deviation_model
        )
    else:
        beta_model = np.sign(exit_plane["beta"]) * math.arccosd(
            critical_mass_flow / exit_plane["d"] / exit_plane["w"] / geometry["A_out"]
        )
    choking_residual = math.cosd(beta_model) - math.cosd(exit_plane["beta"])

    # Restructure critical state dictionary
    inlet_plane = {
        f"critical_{key}_in": val for key, val in critical_state["inlet_plane"].items()
    }
    throat_plane = {
        f"critical_{key}_throat": val
        for key, val in critical_state["throat_plane"].items()
    }
    critical_state = {**inlet_plane, **throat_plane}

    # Return last 3 equations of the Lagrangian gradient (df/dx2+l1*dg1/dx2+l2*dg2/dx2 and g1, g2)
    g = f0[1:]  # The two constraints
    residual_values = np.concatenate((g, np.array([grad, choking_residual])))
    residual_keys = ["m*", "Y*", "L*", "choking"]
    residuals_critical = dict(zip(residual_keys, residual_values))

    return residuals_critical, critical_state


def compute_critical_values(
    x_crit,
    inlet_plane,
    fluid,
    geometry,
    angular_speed,
    critical_state,
    model_options,
    reference_values,
):
    """
    Compute cascade performance at the critical conditions

    This function evaluates the performance of a cascade at its critical operating point defined by:

        1. Critical inlet absolute velocity,
        2. Critical throat relative velocity,
        3. Critical throat entropy.

    Using these variables, the function calculates the critical mass flow rate and residuals of the mass balance and the loss model equations.

    Parameters
    ----------
    x_crit : numpy.ndarray
        Array containing scaled critical variables `[v_in*, w_throat*, s_throat*]`.
    inlet_plane : dict
        Dictionary containing data on the inlet plane for the actual cascade operating condition.
    fluid : object
        A fluid object with methods for thermodynamic property calculations.
    geometry : dict
        Geometric parameters of the cascade.
    angular_speed : float
        Angular speed of the cascade.
    critical_state : dict
        Dictionary to store the critical state information.
    model_options : dict
        Options for the model used in the critical condition evaluation.
    reference_values : dict
        Reference values used in the calculations, including mass flow reference and other parameters.

    Returns
    -------
    numpy.ndarray
        An array containing the computed mass flow at the throat plane at critical state and the residuals
        for mass conservation and loss coefficient error.

    """

    # Define model options
    loss_model = model_options["loss_model"]

    # Load reference values
    mass_flow_ref = reference_values["mass_flow_ref"]
    v0 = reference_values["v0"]
    s_range = reference_values["s_range"]
    s_min = reference_values["s_min"]

    # Load input for critical cascade
    s_in = inlet_plane["s"]
    h0_in = inlet_plane["h0"]
    alpha_in = inlet_plane["alpha"]
    v_in, w_throat, s_throat = (
        x_crit[0] * v0,
        x_crit[1] * v0,
        x_crit[2] * s_range + s_min,
    )

    # Evaluate inlet plane
    critical_inlet_input = {
        "v": v_in,
        "s": s_in,
        "h0": h0_in,
        "alpha": alpha_in,
    }
    critical_inlet_plane = fm.evaluate_cascade_inlet(
        critical_inlet_input, fluid, geometry, angular_speed
    )

    # Evaluate throat plane
    critical_throat_input = {
        "w": w_throat,
        "s": s_throat,
        "beta": np.sign(geometry["gauging_angle"])
        * math.arccosd(geometry["A_throat"] / geometry["A_out"]),
        "rothalpy": critical_inlet_plane["rothalpy"],
    }

    critical_throat_plane, loss_dict = fm.evaluate_cascade_throat(
        critical_throat_input,
        fluid,
        geometry,
        critical_inlet_plane,
        angular_speed,
        model_options["blockage_model"],
        loss_model,
    )

    # Add residuals
    residuals = np.array(
        [
            (critical_inlet_plane["mass_flow"] - critical_throat_plane["mass_flow"])
            / mass_flow_ref,
            critical_throat_plane["loss_error"],
        ]
    )

    # Update critical state dictionary
    critical_state["inlet_plane"] = critical_inlet_plane
    critical_state["throat_plane"] = critical_throat_plane

    output = np.insert(residuals, 0, critical_throat_plane["mass_flow"])

    return output


def compute_critical_jacobian(
    x,
    inlet_plane,
    fluid,
    geometry,
    angular_speed,
    critical_state,
    model_options,
    reference_values,
    f0,
):
    """
    Compute the Jacobian matrix of the compute_critical_values function using finite differences.

    This function approximates the Jacobian of a combined function that includes the mass flow rate value,
    mass balance residual, and loss model evaluation residual at the critical point. It uses forward finite
    difference to approximate the partial derivatives of the Jacobian matrix.

    Parameters
    ----------
    x : numpy.ndarray
        Array of input variables for the `compute_critical_values` function.
    inlet_plane : dict
        Dictionary containing data on the inlet plane for the actual cascade operating condition.
    fluid : object
        A fluid object with methods for thermodynamic property calculations.
    geometry : dict
        Geometric parameters of the cascade.
    angular_speed : float
        Angular speed of the cascade.
    critical_state : dict
        Dictionary to store the critical state information.
    model_options : dict
        Options for the model used in the critical condition evaluation.
    reference_values : dict
        Reference values used in the calculations, including mass flow reference and other parameters.
    f0 : numpy.ndarray
        The function value at x, used for finite difference approximation.

    Returns
    -------
    numpy.ndarray
        The approximated Jacobian matrix of the compute_critical_values function.

    """

    # Define finite difference relative step size
    eps = model_options["rel_step_fd"] * x

    # Approximate problem Jacobian by finite differences
    jacobian = approx_derivative(
        compute_critical_values,
        x,
        method="2-point",
        f0=f0,
        abs_step=eps,
        args=(
            inlet_plane,
            fluid,
            geometry,
            angular_speed,
            critical_state,
            model_options,
            reference_values,
        ),
    )

    return jacobian


def evaluate_cascade_isentropic_throat(
    choking_input,
    inlet_plane,
    exit_plane,
    fluid,
    geometry,
    angular_speed,
    model_options,
    reference_values,
):
    r"""

    Calculate condition for choking and evaluate wheter or not the cascade is choked, based on the `evaluate_cascade_isentropic_throat` choking model.

    This choking model evaluates the cascade throat and checks if the throat mach number exceed unity. The throat is isentropic from inlet to throat.
    The exit flow angle is determined by ensuring that the mach at throat mach number equals the exit for subconic condition, and unity for supersonic conditions.

    Parameters
    ----------
    choking_input : dict
        Dictionary containing scaled input parameters required for selected choking model. Required items are:

        - `w*_throat` : throat velocity.
    inlet_plane : dict
        Dictionary containing data on the inlet plane for the actual cascade operating condition.
    exit_plane : dict
        Dictionary containing data on the exit plane for the actual cascade operating condition.
    fluid : object
        A fluid object with methods for thermodynamic property calculations.
    geometry : dict
        Geometric parameters of the cascade.
    angular_speed : float
        Angular speed of the cascade.
    model_options : dict
        Options for the model used to evaluate choking.
    reference_values : dict
        Reference values used in the calculation, including the reference mass flow rate.

    Returns
    -------
    dict
        Dictionary containing the residuals of choking model:

        - `m*`: Mass flow rate residual.
        - `choking`: Choking residual.
    dict
        Dictionary satisfying the required elements for critical state dict.

    """

    # Rename variables
    loss_model = model_options["loss_model"]
    blockage = model_options["blockage_model"]
    deviation_model = model_options["deviation_model"]
    A_throat = geometry["A_throat"]
    A_out = geometry["A_out"]
    v0 = reference_values["v0"]
    angle_range = reference_values["angle_range"]
    angle_min = reference_values["angle_min"]

    # Evaluate throat
    cascade_throat_input = {
        "w": choking_input["w*_throat"] * v0,
        "s": inlet_plane["s"],
        "beta": np.sign(exit_plane["beta"])
        * math.arccosd(geometry["A_throat"] / geometry["A_out"]),
        "rothalpy": inlet_plane["rothalpy"],
    }
    throat_plane, loss_dict = fm.evaluate_cascade_throat(
        cascade_throat_input,
        fluid,
        geometry,
        inlet_plane,
        angular_speed,
        blockage,
        loss_model,
    )

    # Evaluate critical mach
    choking_residual = throat_plane["Ma_rel"] - min(exit_plane["Ma_rel"], 1)

    # Evaluate resiudals
    residual_values = np.array(
        [
            (inlet_plane["mass_flow"] - throat_plane["mass_flow"])
            / reference_values["mass_flow_ref"],
            choking_residual,
        ]
    )
    residual_keys = ["m*", "choking"]
    residuals_critical = dict(zip(residual_keys, residual_values))

    # Define output values
    throat_plane = {f"{key}_throat": val for key, val in throat_plane.items()}

    return residuals_critical, throat_plane



# Highlight exception messages
# https://stackoverflow.com/questions/25109105/how-to-colorize-the-output-of-python-errors-in-the-gnome-terminal/52797444#52797444
try:
    import IPython.core.ultratb
except ImportError:
    # No IPython. Use default exception printing.
    pass
else:
    import sys

    sys.excepthook = IPython.core.ultratb.ColorTB()

import toml

# Import submodules
from .math import *
from .plot_functions import *
from .config_validation import *

# Import subpackages
from .pysolver_view import *
from .properties import *
from .cycles import *
from .utilities import *

from .axial_turbine import *
from .axial_turbine.performance_analysis import *
from .axial_turbine.design_optimization import *

# Set plot options
set_plot_options()


import os
import toml

BREAKLINE = 80 * "-"


def print_banner():
    """Prints a banner."""
    banner = """
        ______ __  __ ____   ____   ____   ______ __    ____  _       __
       /_  __// / / // __ \ / __ ) / __ \ / ____// /   / __ \| |     / /
        / /  / / / // /_/ // __  |/ / / // /_   / /   / / / /| | /| / / 
       / /  / /_/ // _, _// /_/ // /_/ // __/  / /___/ /_/ / | |/ |/ /   
      /_/   \____//_/ |_|/_____/ \____//_/    /_____/\____/  |__/|__/  
    """
    print(BREAKLINE)
    print(banner)


def print_package_info():
    """Prints package information."""
    try:
        pyproject_path = find_pyproject_toml()
        with open(pyproject_path, "r") as f:
            pyproject = toml.load(f)

        poetry_config = pyproject.get("tool", {}).get("poetry", {})
        info = f""" Version:       {poetry_config.get("version")}
 Repository:    {poetry_config.get("repository")}
 Documentation: {poetry_config.get("documentation")}"""
        print_banner()
        print(BREAKLINE)
        print(info)
    except FileNotFoundError:
        print("Error: pyproject.toml file not found.")


def find_pyproject_toml():
    """Finds the path to the pyproject.toml file."""
    current_script_path = os.path.abspath(__file__)
    current_script_dir = os.path.dirname(current_script_path)
    project_root_dir = os.path.dirname(current_script_dir)
    pyproject_path = os.path.join(project_root_dir, "pyproject.toml")
    return pyproject_path





from . import loss_model_benner as br
from . import loss_model_kacker_okapuu as ko
from . import loss_model_moustapha as mo
from .. import utilities as utils

# List of valid options
LOSS_MODELS = [
    "kacker_okapuu",
    "moustapha",
    "benner",
    "isentropic",
    "custom",
]
"""
Available loss models.
"""

LOSS_COEFFICIENTS = [
    "stagnation_pressure",
]
"""
Available loss coefficients.
"""

# Keys that the output loss dictionary must have
KEYS_LOSSES = [
    "loss_definition",
    "loss_error",
    "loss_total",
    "loss_profile",
    "loss_clearance",
    "loss_secondary",
    "loss_trailing",
    "loss_incidence",
]


def evaluate_loss_model(loss_model_options, input_parameters):
    """
    Calculate loss coefficient based on the selected loss model.

    The avaialble loss models are:

    - `kacker_okapuu` : loss model according to :cite:`kacker_mean_1982`.
    - `moustapha` : loss model according to :cite:`moustapha_improved_1990`.
    - `benner` : loss model according to :cite:`benner_influence_1997`, :cite:`benner_empirical_2006-1` and :cite:`benner_empirical_2006`.
    - `isentropic` : Loss coefficient set to zero (isentropic).
    - `custom` :  Constant loss coefficent according to user. Require `loss_model_options["custom_value"]`.

    Parameters
    ----------
    loss_model_options : dict
        Options for the loss calculation.
    input_parameters : dict
        Input parameters required for loss model calculation.

    Returns:
    dict
        A dictionary containing loss components.
    """

    # Function mappings for each loss model
    loss_model_functions = {
        LOSS_MODELS[0]: ko.compute_losses,
        LOSS_MODELS[1]: mo.compute_losses,
        LOSS_MODELS[2]: br.compute_losses,
        LOSS_MODELS[3]: lambda _: {
            "loss_profile": 0.0,
            "loss_incidence": 0.0,
            "loss_trailing": 0.0,
            "loss_secondary": 0.0,
            "loss_clearance": 0.0,
            "loss_total": 0.0,
        },
        LOSS_MODELS[4]: lambda input_parameters: {
            "loss_profile": 0.0,
            "loss_incidence": 0.0,
            "loss_trailing": 0.0,
            "loss_secondary": 0.0,
            "loss_clearance": 0.0,
            "loss_total": loss_model_options["custom_value"],
        },
    }

    # Evaluate loss model
    model = loss_model_options["model"]
    if model in loss_model_functions:
        loss_dict = loss_model_functions[model](input_parameters)
    else:
        options = ", ".join(f"'{k}'" for k in LOSS_MODELS)
        raise ValueError(f"Invalid loss model '{model}'. Available options: {options}")

    # Apply tuning factors (empty dict if not provided)
    tuning_factors = loss_model_options.get("tuning_factors", {})
    tuning_factors = {f"loss_{key}": value for key, value in tuning_factors.items()}
    apply_tuning_factors(loss_dict, tuning_factors)

    # Compute the loss coefficient according to definition
    loss_coeff = loss_model_options["loss_coefficient"]
    if loss_coeff == "stagnation_pressure":
        # TODO add other loss coefficient definitions
        p0rel_in = input_parameters["flow"]["p0_rel_in"]
        p0rel_out = input_parameters["flow"]["p0_rel_out"]
        p_out = input_parameters["flow"]["p_out"]
        Y_definition = (p0rel_in - p0rel_out) / (p0rel_out - p_out)

    else:
        options = ", ".join(f"'{k}'" for k in LOSS_COEFFICIENTS)
        raise ValueError(
            f"Invalid loss coefficient '{loss_coeff}'. Available options: {options}"
        )

    # Compute loss coefficient error
    loss_dict["loss_error"] = Y_definition - loss_dict["loss_total"]

    return loss_dict


def apply_tuning_factors(loss_dict, tuning_factors):
    """
    Apply tuning factors to the loss model

    The tuning factors are multiplied with their associated loss component.

    Parameters
    ----------
    loss_dict : dict
        A dictionary containing loss components.
    tuning_factors : dict
        A dictionary containing the multiplicative tuning factors.

    Raises:
        KeyError: If a key from `tuning_factors` is not found in `loss_dict`.
    """

    for key, factor in tuning_factors.items():
        if key not in loss_dict:
            raise KeyError(f"Tuning factor key '{key}' not found in loss dictionary.")
        loss_dict[key] *= factor

    return loss_dict



import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt


def load_data(filename):
    """
    Load performance data from an Excel file.

    This function imports an Excel file containing performance parameters. It reads the data into a dictionary
    named 'performance_data' where each key corresponds to a sheet name in the Excel file, and each value is
    a pandas DataFrame containing the data from that sheet. The data is rounded to avoid precision loss when
    loading from Excel.

    The excel file must contain the following sheets:

        - `operation point`
        - `plane`
        - `cascade`
        - `overall`

    Parameters
    ----------
    filename : str
        The name of the Excel file containing performance parameters.

    Returns
    -------
    dict
        A dictionary containing performance data, where keys are sheet names and values are DataFrames.

    """

    # Read excel file
    performance_data = pd.read_excel(
        filename,
        sheet_name=[
            "operation point",
            "plane",
            "cascade",
            # "stage",
            "overall",
        ],
    )

    # Round off to ignore precision loss by loading data from excel
    for key, df in performance_data.items():
        performance_data[key] = df.round(10)

    return performance_data


def plot_lines(
    performance_data,
    x_key,
    y_keys,
    subsets=None,
    fig=None,
    ax=None,
    xlabel="",
    ylabel="",
    title="",
    labels=None,
    filename=None,
    outdir="figures",
    stack=False,
    color_map="viridis",
    colors=None,
    close_fig=False,
    save_figs=False,
    linestyles=None,
    save_formats=[".png"],
    legend_loc="best",
):
    """
    Plot lines from performance data.

    This function plots lines from performance data. It supports plotting multiple lines on the same axes
    with customizable labels, colors, and linestyles. The resulting plot can be saved to a file if desired.

    Parameters
    ----------
    performance_data : DataFrame
        The performance data to plot.
    x_key : str
        Name of the column in performance_data to be used as x-axis values.
    y_keys : list of str
        Names of the columns in performance_data to be plotted.
    subsets : list, optional
        Name and value of subsets to plot from performance_data. First instance should be a string representing the column name, while the
        remaining elemnts represet the values defining the subset. Default is None.
    fig : matplotlib.figure.Figure, optional
        An existing figure object. If None, a new figure is created.
    ax : matplotlib.axes.Axes, optional
        An existing axes object. If None, a new axes is created.
    xlabel : str, optional
        The label for the x-axis.
    ylabel : str, optional
        The label for the y-axis.
    title : str, optional
        The title of the plot.
    labels : list of str, optional
        Labels for the plotted lines. Default is None.
    filename : str, optional
        The filename for saving the figure. Default is None.
    outdir : str, optional
        The directory where figures should be saved. Default is 'figures'.
    stack : bool, optional
        Whether to stack the plotted lines. Default is False.
    color_map : str, optional
        The colormap used for the lines. Default is 'viridis'.
    colors : list of str or colors, optional
        Colors for the plotted lines. Default is None.
    close_fig : bool, optional
        Whether to close the figure after plotting. Default is False.
    save_figs : bool, optional
        Whether to save the figure. Default is False.
    linestyles : list of str, optional
        Linestyles for the plotted lines. Default is None.
    save_formats : list of str, optional
        File formats for saving the figure. Default is ['.png'].
    legend_loc : str, optional
        Location for the legend. Default is 'best'.

    Returns
    -------
    matplotlib.figure.Figure
        The figure object.
    matplotlib.axes.Axes
        The axes object.

    """

    # Create figure if not provided
    if fig is None or ax is None:
        fig, ax = plt.subplots(figsize=(6.4, 4.8))

    # Specify title and axes labels
    ax.set_title(title)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    # Get data
    if subsets == None:
        x = get_lines(performance_data, x_key)
        y = get_lines(performance_data, y_keys)
    else:
        x = get_lines(performance_data, x_key, subsets=subsets)
        y = get_lines(performance_data, y_keys, subsets=subsets)

    # Get colors
    if colors == None:
        colors = plt.get_cmap(color_map)(np.linspace(0.2, 1.0, len(y)))

    # Get labels
    if labels == None:
        if subsets is not None and len(subsets) >2:
            labels = [f"{subsets[0]} = {subsets[i+1]}" for i in range(len(subsets) - 1)]
        elif len(y_keys) > 1:
            labels = [f"{y_keys[i]}" for i in range(len(y))]
        else:
            labels = y_keys

    # Get linestyles
    if linestyles == None:
        linestyles = ["-"] * len(y)

    # Plot figure
    if subsets is not None and len(y_keys) == 1:
        for i in range(len(y)):
            ax.plot(x[i], y[i], label=labels[i], color=colors[i], linestyle=linestyles[i])
    elif subsets is not None and len(subsets) == 2:
        if stack == True:
            ax.stackplot(x[0], y, labels=labels, colors=colors)
            # Add edges by overlaying lines
            y_arrays = [series.values for series in y]
            cumulative_y = np.cumsum(y_arrays, axis=0)
            for i, series in enumerate(y_arrays):
                if i == 0:
                    ax.plot(x[0], series, color="black", linewidth=0.5)
                ax.plot(x[0], cumulative_y[i], color="black", linewidth=0.5)
        else:
            for i in range(len(y)):
                ax.plot(
                    x[0], y[i], label=labels[i], color=colors[i], linestyle=linestyles[i]
                )
    elif subsets is None:
        if stack == True:
            ax.stackplot(x, y, labels=labels, colors=colors)
            # Add edges by overlaying lines
            y_arrays = [series.values for series in y]
            cumulative_y = np.cumsum(y_arrays, axis=0)
            for i, series in enumerate(y_arrays):
                if i == 0:
                    ax.plot(x, series, color="black", linewidth=0.5)
                ax.plot(x, cumulative_y[i], color="black", linewidth=0.5)
        else:
            for i in range(len(y)):
                ax.plot(x, y[i], label=labels[i], color=colors[i], linestyle=linestyles[i])
    else:
        raise Exception("x_keys and y_keys are not well defined")

    # Set margins to zero
    ax.margins(x=0.01, y=0.01)

    # Add legend
    if len(y) > 1:
        ax.legend(loc=legend_loc)
    fig.tight_layout(pad=1, w_pad=None, h_pad=None)

    # Create output directory if it does not exist
    if not os.path.exists(outdir):
        os.makedirs(outdir)

    # Save figure
    if save_figs:
        base_filename = os.path.join(outdir, filename)
        savefig_in_formats(fig, base_filename, formats=save_formats)

    if close_fig:
        plt.close(fig)

    return fig, ax


def get_lines(performance_data, column_name, subsets=None):
    """
    Retrieve lines of data from the specified column in `performance_data`.

    This function returns a list of array from the specified column(s) of the performance data.
    If no subset is specified, it returns lines covering all rows in `performance_data`. If a subset
    is specified, it returns lines covering only the rows that match the specified subset.

    Parameters
    ----------
    performance_data : DataFrame
        The DataFrame containing the performance data.
    column_name : str or list of str
        The name(s) of the column(s) from which to retrieve data.
    subsets : list, optional
        Name and value of subsets to get data from performance_data. First instance should be a string representing the column name, while the
        remaining elemnts represet the values defining the subset. Default is None.

    Returns
    -------
    list
        A list of arrays from the specified column(s) of performance_data.

    """

    # Get lines covering all rows in performance_data
    if subsets == None:
        # Get single line
        if isinstance(column_name, str):
            return get_column(performance_data, column_name)
        # Get several columns
        else:
            lines = []
            for column in column_name:
                lines.append(get_column(performance_data, column))
            return lines

    subsets[1:] = [round(value, 10) for value in subsets[1:]]
    # Get lines covering given subset
    lines = []
    for val in subsets[1:]:
        if isinstance(column_name, str):
            indices = get_subset(performance_data, subsets[0], val)
            lines.append(get_column(performance_data, column_name)[indices])
        else:
            for column in column_name:
                indices = get_subset(performance_data, subsets[0], val)
                lines.append(get_column(performance_data, column)[indices])

    return lines


def find_column(performance_data, column_name):
    """
    Find the sheet containing the specified column in `performance_data`.

    This function searches through all sheets in `performance_data` to find the one containing the specified column.
    If the column is found, it returns the name of the sheet. If the column is not found, it raises an exception.

    Parameters
    ----------
    performance_data : dict
        A dictionary containing sheets of performance data, where keys are sheet names and values are DataFrames.
    column_name : str
        The name of the column to find.

    Returns
    -------
    str
        The name of the sheet containing the specified column.

    Raises
    ------
    Exception
        If the specified column is not found in any sheet of performance_data.

    """

    for key in performance_data.keys():
        if any(element == column_name for element in performance_data[key].columns):
            return key

    raise Exception(f"Could not find column {column_name} in performance_data")


def get_column(performance_data, column_name):
    """
    Retrieve a column of data from `performance_data`.

    This function retrieves the specified column of data from `performance_data` by finding the sheet containing
    the column using the `find_column` function. It then returns the column as a pandas Series.

    Parameters
    ----------
    performance_data : dict
        A dictionary containing sheets of performance data, where keys are sheet names and values are DataFrames.
    column_name : str
        The name of the column to retrieve.

    Returns
    -------
    pandas.Series
        The column of data corresponding to column_name.

    """
    sheet = find_column(performance_data, column_name)

    return performance_data[sheet][column_name]


def get_subset(performance_data, column_name, row_value):
    """
    Retrieve the index of rows in `performance_data` where `column_name` equals `row_value`.

    This function retrieves the index of rows in `performance_data` where the specified column (`column_name`)
    equals the specified value (`row_value`). It first finds the sheet containing the column using the
    `find_column` function, then returns the index of rows where the column has the specified value.

    Parameters
    ----------
    performance_data : dict
        A dictionary containing sheets of performance data, where keys are sheet names and values are DataFrames.
    column_name : str
        The name of the column to search for `row_value`.
    row_value : object
        The value to match in `column_name`.

    Returns
    -------
    pandas.Index
        The index of rows where `column_name` equals `row_value`.

    """
    sheet = find_column(performance_data, column_name)

    return performance_data[sheet][
        performance_data[sheet][column_name] == row_value
    ].index


def savefig_in_formats(
    fig, path_without_extension, formats=[".png", ".svg", ".pdf", ".eps"]
):
    """
    Save a given matplotlib figure in multiple file formats.

    Parameters
    ----------
    fig : matplotlib.figure.Figure
        The figure object to be saved.
    path_without_extension : str
        The full path to save the figure excluding the file extension.
    formats : list of str, optional
        A list of string file extensions to specify which formats the figure should be saved in.
        Default is ['.png', '.svg', '.pdf', ".eps"].

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> fig, ax = plt.subplots()
    >>> ax.plot([0, 1], [0, 1])
    >>> save_fig_in_formats(fig, "/path/to/figure/filename")

    This will save the figure as "filename.png", "filename.svg", and "filename.pdf" in the "/path/to/figure/" directory.
    """
    for ext in formats:
        fig.savefig(f"{path_without_extension}{ext}", bbox_inches="tight")


def plot_axial_radial_plane(geometry):
    r"""
    Plot the turbine geometry in an axial-radial plane.

    This function plots the turbine geometry in an axial-radial plane. It takes the turbine geometry data
    as input, including the radii at the inner and outer hub and tip, the number of cascades, and the axial
    chord lengths. It then constructs and displays the plot, which represents the turbine blades in the
    axial-radial plane.

    Parameters
    ----------
    geometry : dict
        A dictionary containing turbine geometry data including:

        - `radius_hub_in` (array-like) : Inner hub radii at each cascade.
        - `radius_hub_out` (array-like) : Outer hub radii at each cascade.
        - `radius_tip_in` (array-like) : Inner tip radii at each cascade.
        - `radius_tip_out` (array-like) : Outer tip radii at each cascade.
        - `number_of_cascades` (int) : Number of cascades in the turbine.
        - `axial_chord` (array-like) : Axial chord lengths at each cascade.

    Returns
    -------
    matplotlib.figure.Figure
        The generated figure.
    matplotlib.axes.Axes
        The generated axes.

    """

    # Load data
    radius_hub_in = geometry["radius_hub_in"]
    radius_hub_out = geometry["radius_hub_out"]
    radius_tip_in = geometry["radius_tip_in"]
    radius_tip_out = geometry["radius_tip_out"]
    number_of_cascades = geometry["number_of_cascades"]
    axial_chord = geometry["axial_chord"]
    tip_clearance = geometry["tip_clearance"]

    # Define unspecified geometrical parameters
    ct = 0.05  # Thickness of casing relative to maximum tip radius
    cl = 1.2  # Length of casing to turbine
    cs = 0.1  # Axial spacing between cascades relative to max axial chord

    # Get x-points
    dx = cs * max(axial_chord)
    x = np.array([])
    for i in range(len(axial_chord)):
        x = np.append(x, np.sum(axial_chord[0 : i + 1]) + i * dx)
        x = np.append(x, np.sum(axial_chord[0 : i + 1]) + (i + 1) * dx)
    x[1:] = x[0:-1]
    x[0] = 0

    # Get tip and hub points
    y_hub = np.array(
        [val for pair in zip(radius_hub_in, radius_hub_out) for val in pair]
    )
    y_tip = np.array(
        [val for pair in zip(radius_tip_in, radius_tip_out) for val in pair]
    )

    # Initialize figure and axis object
    fig, ax = plt.subplots()

    # Plot upper casing
    tip_clearance_extended = np.array(
        [val for pair in zip(tip_clearance, tip_clearance) for val in pair]
    )  # Extend tip clearance vector
    casing_thickness = max(y_tip) * ct
    x_casing = x
    x_casing = np.insert(x_casing, 0, x[0] - (cl - 1) * x[-1] / 2)
    x_casing = np.append(x_casing, x[-1] + (cl - 1) * x[-1] / 2)

    y_casing_lower = y_tip + tip_clearance_extended
    y_casing_lower = np.insert(y_casing_lower, 0, y_casing_lower[0])
    y_casing_lower = np.append(y_casing_lower, y_casing_lower[-1])
    y_casing_upper = np.ones(len(x_casing)) * (max(y_tip) + casing_thickness)

    ax.plot(x_casing, y_casing_lower, "k")
    ax.plot(x_casing, y_casing_upper, "k")
    ax.plot([x_casing[0], x_casing[0]], [y_casing_lower[0], y_casing_upper[0]], "k")
    ax.plot([x_casing[-1], x_casing[-1]], [y_casing_lower[-1], y_casing_upper[-1]], "k")
    ax.fill_between(x_casing, y_casing_lower, y_casing_upper, color="0.5")

    # Plot lower casing
    tip_clearance_extended = np.array(
        [
            val
            for pair in zip(np.flip(tip_clearance), np.flip(tip_clearance))
            for val in pair
        ]
    )  # Extend tip clearance vector
    y_casing_upper = y_hub - tip_clearance_extended
    y_casing_upper = np.insert(y_casing_upper, 0, y_casing_upper[0])
    y_casing_upper = np.append(y_casing_upper, y_casing_upper[-1])
    y_casing_lower = y_casing_upper - casing_thickness

    ax.plot(x_casing, y_casing_lower, "k")
    ax.plot(x_casing, y_casing_upper, "k")
    ax.plot([x_casing[0], x_casing[0]], [y_casing_lower[0], y_casing_upper[0]], "k")
    ax.plot([x_casing[-1], x_casing[-1]], [y_casing_lower[-1], y_casing_upper[-1]], "k")
    ax.fill_between(x_casing, y_casing_lower, y_casing_upper, color="0.5")

    # Plot cascades
    for i in range(number_of_cascades):
        ax.plot(x[i * 2 : (i + 1) * 2], y_tip[i * 2 : (i + 1) * 2], "k")  # Plot tip
        ax.plot(x[i * 2 : (i + 1) * 2], y_hub[i * 2 : (i + 1) * 2], "k")  # Plot hub
        ax.plot(
            [x[i * 2 : (i + 1) * 2][0], x[i * 2 : (i + 1) * 2][0]],
            [y_hub[i * 2 : (i + 1) * 2][0], y_tip[i * 2 : (i + 1) * 2][0]],
            "k",
        )  # Plot inlet vertical line
        ax.plot(
            [x[i * 2 : (i + 1) * 2][1], x[i * 2 : (i + 1) * 2][1]],
            [y_hub[i * 2 : (i + 1) * 2][1], y_tip[i * 2 : (i + 1) * 2][1]],
            "k",
        )  # Plot exit vertical line
        ax.fill_between(
            x[i * 2 : (i + 1) * 2],
            y_hub[i * 2 : (i + 1) * 2],
            y_tip[i * 2 : (i + 1) * 2],
            color="1",
        )  # Fill cascade with color

    # Plot axis of rotation
    plt.annotate(
        "",
        xy=(x[-1], 0),
        xytext=(0, 0),
        arrowprops=dict(facecolor="black", arrowstyle="-|>"),
    )

    ax.grid(False)
    ax.tick_params(which="both", direction="out", right=False)
    ax.tick_params(which="minor", left=False)
    # ax.set_yticks([])
    ax.set_xticks([])
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["bottom"].set_visible(False)
    # ax.spines['left'].set_visible(False)
    ax.set_xlabel("Axis of rotation")
    ax.set_ylabel("Radius")
    ax.set_aspect("equal")
    lims = ax.get_ylim()
    ax.set_ylim([0, lims[-1]])
    plt.show()

    return fig, ax


def plot_velocity_triangles(plane):
    """
    Plot velocity triangles for each stage of a turbine.

    Parameters
    ----------
    plane : dict
        A dictionary containing velocity components for each plane of the turbine.
        The dictionary should have the following keys:
        - 'v_m': Meridional velocities.
        - 'v_t': Tangential velocities.
        - 'w_m': Meridional velocities of relative motion.
        - 'w_t': Tangential velocities of relative motion.

    Returns
    -------
    matplotlib.figure.Figure
        The matplotlib figure object containing the velocity triangle plot.

    matplotlib.axes._subplots.AxesSubplot
        The matplotlib axis object containing the velocity triangle plot.

    """

    # Load variables
    v_ms = plane["v_m"]
    v_ts = plane["v_t"]
    w_ms = plane["w_m"]
    w_ts = plane["w_t"]

    # Get maximum an minimum tangential velocities
    max_w_t = max(np.array([val for pair in zip(v_ts, w_ts) for val in pair]))
    min_w_t = min(np.array([val for pair in zip(v_ts, w_ts) for val in pair]))

    # Nameing of planes (for marking)
    plane_name = ["inlet", "exit"]

    # Tuning options
    dx = 0.1  # Spacing between triangles
    hs = 10  # Horizontal spacing between line and symbol (absolute)
    start = 0  # Starting point in y direction
    fontsize = 12  # Fontsize

    # Get rotor indexes
    i_rotor = np.array(
        [val for pair in zip(plane.index[2::4], plane.index[3::4]) for val in pair]
    )

    # initialize figure and axis
    fig, ax = plt.subplots()

    # Plot velocity triangles
    for i in range(len(plane)):

        v_m = v_ms[i]
        v_t = v_ts[i]

        w_m = w_ms[i]
        w_t = w_ts[i]

        # Define cascade name
        cascade = "Stator"  # Changed if rotor (see below)

        # Plot absolute velocity
        ax.plot(
            [0, 0], [start, start - v_m], color="k", linestyle=":"
        )  # Plot meridional velocity
        ax.plot(
            [0, v_t], [start - v_m, start - v_m], color="k", linestyle=":"
        )  # Plot tangntial velocity
        ax.plot([0, v_t], [start, start - v_m], "k")  # Plot velocity
        x_mid = v_t / 2
        y_mid = (start + start - v_m) / 2
        plt.annotate(
            "",
            xy=(x_mid, y_mid),
            xytext=(0, start),
            arrowprops=dict(facecolor="black", arrowstyle="-|>"),
        )  # Arrow for absolute velocity
        sign = -1 if x_mid < 0 else 1
        ax.text(
            x_mid + hs * sign,
            y_mid,
            f"$v$",
            fontsize=fontsize,
            ha="center",
            va="bottom",
            color="k",
        )  # Symbol for absolute velocity

        # Plot relative velocity
        if i in i_rotor:

            cascade = "Rotor"
            ax.plot(
                [0, w_t], [start - w_m, start - w_m], color="k", linestyle=":"
            )  # Plot tangntial velocity
            ax.plot([0, w_t], [start, start - w_m], color="k")  # Plot velocity
            ax.plot([w_t, v_t], [start - w_m, start - w_m], "k")  # Plot blade speed
            x_mid = w_t / 2
            y_mid = (start + start - w_m) / 2
            plt.annotate(
                "",
                xy=(x_mid, y_mid),
                xytext=(0, start),
                arrowprops=dict(facecolor="black", arrowstyle="-|>"),
            )  # Arrow for relative velocity
            sign = -1 if x_mid < 0 else 1
            ax.text(
                x_mid + hs * np.sign(x_mid),
                y_mid,
                f"$w$",
                fontsize=fontsize,
                ha="center",
                va="bottom",
                color="k",
            )  # Symbol for relative velocity
            x_mid = (w_t + v_t) / 2
            y_mid = start - w_m
            plt.annotate(
                "",
                xy=(x_mid, y_mid),
                xytext=(w_t, start - w_m),
                arrowprops=dict(facecolor="k", arrowstyle="-|>"),
            )  # Arrow for blade speed
            ax.text(
                x_mid,
                y_mid,
                f"$u$",
                fontsize=fontsize,
                ha="center",
                va="bottom",
                color="k",
            )  # Symbol for blade speed

        # Add explanatory text
        if abs(v_t) >= abs(w_t):
            tangential_component = v_t
        else:
            tangential_component = w_t
        if tangential_component < 0:
            ax.text(
                max_w_t / 2,
                start - v_m / 2,
                f"{cascade} {plane_name[i%2]}",
                fontsize=fontsize,
                ha="center",
                va="center",
                color="k",
            )
        else:
            ax.text(
                min_w_t / 2,
                start - v_m / 2,
                f"{cascade} {plane_name[i%2]}",
                fontsize=fontsize,
                ha="center",
                va="center",
                color="k",
            )

        # Plot line diciding the different triangles
        ax.plot(
            [min_w_t, max_w_t], [start - w_m - dx * v_m, start - w_m - dx * v_m], "k:"
        )

        # Update starting point for the next velocity triangle
        start = start - w_m - 2 * dx * v_m

    # Set plot options
    ax.grid(False)
    ax.set_yticks([])
    ax.set_xticks([])
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["bottom"].set_visible(False)
    ax.spines["left"].set_visible(False)
    plt.show()

    return fig, ax



import numpy as np
import pandas as pd
import CoolProp as cp
from scipy.linalg import solve
from scipy.optimize._numdiff import approx_derivative

from .. import math
from .. import utilities as utils
from . import loss_model as lm
from . import deviation_model as dm
from . import choking_model as cm

# List of valid options
BLOCKAGE_MODELS = ["flat_plate_turbulent"]


def evaluate_axial_turbine(
    variables,
    boundary_conditions,
    geometry,
    fluid,
    model_options,
    reference_values,
):
    """
    Compute the performance of an axial turbine by evaluating a series of cascades.

    The function iterates through each cascade, accumulating data on both performance and residuals used to evaluate the physicality of the flow.
    Between each cascade, the space between the cascades is modelled to give the input to the next cascade.

    The results are organized into a dictionary that includes:

    - `cascade`: Contains data specific to each cascade in the series, including loss coefficients and critical conditions.
    - `plane`: Contains data specific to each flow station, including thermodynamic properties and velocity triangles.
    - `stage`: Contains data specific to each turbine stage, including degree of reaction.
    - `geometry`: Contains data on the turbine geometry.
    - `overall`: Summarizes the overall performance of the turbine, including mass flow rate, efficiency and power output.
    - `reference_values`: Contain supplementary information.
    - `residuals`: Summarizes the mismatch in the nonlinear system of equations used to model the turbine.

    .. note::

        The output of this function can only be regarded as a physically meaningful solution when the equations are fully closed (i.e., all residuals are close to zero).
        Therefore, this function is intended to be used within a root-finding or optimization algorithm that systematically adjusts the input variables to drive the residuals to zero and close the system of equations.

    Parameters
    ----------
    variables : dict
        Dictionary containing flow variable for the cascades.
    boundary_conditions : dict
        Dictionary containing boundary conditions for the series of cascades.
    geometry : dict
        Dictionary containing geometric parameters of the series of cascades.
    fluid : object
        A fluid object with methods for thermodynamic property calculations.
    model_options : dict
        Dictionary containing various model options.
    reference_values : dict
        Dictionary containing reference values for scaling.

    Returns
    -------
    results : dict
        Dictionary containing the evaluated results, including planes, cascades, stage, overall,
        and geometry information.

    """

    # Load variables
    number_of_cascades = geometry["number_of_cascades"]
    h0_in = boundary_conditions["h0_in"]
    s_in = boundary_conditions["s_in"]
    alpha_in = boundary_conditions["alpha_in"]
    angular_speed = boundary_conditions["omega"]

    # Load reference_values
    v0 = reference_values["v0"]
    s_range = reference_values["s_range"]
    s_min = reference_values["s_min"]
    angle_range = reference_values["angle_range"]
    angle_min = reference_values["angle_min"]

    # Initialize results structure
    results = {
        "plane": pd.DataFrame(),
        "cascade": pd.DataFrame(),
        "stage": {},
        "overall": {},
        "geometry": geometry,
        "reference_values": reference_values,
        "boundary_conditions": boundary_conditions,
    }

    # initialize residual arrays
    residuals = {}

    # Rename turbine inlet velocity
    v_in = variables["v_in"] * v0

    for i in range(number_of_cascades):
        # Update angular speed
        angular_speed_cascade = angular_speed * (i % 2)

        # update geometry for current cascade
        geometry_cascade = {
            key: values[i]
            for key, values in geometry.items()
            if key not in ["number_of_cascades", "number_of_stages"]
        }

        # Rename variables
        cascade = "_" + str(i + 1)
        w_out = variables["w_out" + cascade] * v0
        s_out = variables["s_out" + cascade] * s_range + s_min
        beta_out = variables["beta_out" + cascade] * angle_range + angle_min

        # Evaluate current cascade
        cascade_inlet_input = {
            "h0": h0_in,
            "s": s_in,
            "alpha": alpha_in,
            "v": v_in,
        }
        cascade_exit_input = {
            "w": w_out,
            "beta": beta_out,
            "s": s_out,
        }

        choking_input = {
            key.replace(cascade, ""): val
            for key, val in variables.items()
            if ("*" and cascade in key) or key == "v*_in"
        }

        cascade_residuals = evaluate_cascade(
            cascade_inlet_input,
            cascade_exit_input,
            choking_input,
            fluid,
            geometry_cascade,
            angular_speed_cascade,
            results,
            model_options,
            reference_values,
        )

        # Add cascade residuals to residual array
        cascade_residuals = utils.add_string_to_keys(cascade_residuals, f"_{i+1}")
        residuals.update(cascade_residuals)

        # Calculate input of next cascade
        if i != number_of_cascades - 1:
            (
                h0_in,
                s_in,
                alpha_in,
                v_in,
            ) = evaluate_cascade_interspace(
                results["plane"]["h0"].values[-1],
                results["plane"]["v_m"].values[-1],
                results["plane"]["v_t"].values[-1],
                results["plane"]["d"].values[-1],
                geometry_cascade["radius_mean_out"],
                geometry_cascade["A_out"],
                results["plane"]["blockage"].values[-1],
                geometry["radius_mean_in"][i + 1],
                geometry["A_in"][i + 1],
                fluid,
            )

    # Add exit pressure error to residuals
    p_calc = results["plane"]["p"].values[-1]
    p_error = (p_calc - boundary_conditions["p_out"]) / boundary_conditions["p0_in"]
    residuals["p_out"] = p_error

    # Additional calculations
    loss_fractions = compute_efficiency_breakdown(results)
    results["cascade"] = pd.concat([results["cascade"], loss_fractions], axis=1)

    # Compute stage performance
    results["stage"] = pd.DataFrame(compute_stage_performance(results))

    # Compute overall perfrormance
    results["overall"] = pd.DataFrame([compute_overall_performance(results, geometry)])

    # Store all residuals for export
    results["residuals"] = residuals

    # Store the input variables
    results["independent_variables"] = variables

    # Retain only variables defined per cascade
    geom_cascades = {
        key: value
        for key, value in geometry.items()
        if len(utils.ensure_iterable(value)) == number_of_cascades
    }
    results["geometry"] = pd.DataFrame([geom_cascades])

    return results


def evaluate_cascade(
    cascade_inlet_input,
    cascade_exit_input,
    choking_input,
    fluid,
    geometry,
    angular_speed,
    results,
    model_options,
    reference_values,
):
    """
    Evaluate the performance of a cascade configuration.

    This function evaluates the cascade performance by considering inlet, throat, and exit planes.
    It also determines the condition for choking according to the selected choking model, and evaluates if the cascade is choked or not. The results are stored in a dictionary.
    The function returns a dictionary of residuals that includes the mass balance error at both the inlet and exit, loss coefficient errors,
    and the residuals related to the critical state and choking condition.

    Parameters
    ----------
    cascade_inlet_input : dict
        Input conditions at the cascade inlet.
    cascade_throat_input : dict
        Input conditions at the cascade throat.
    cascade_exit_input : dict
        Input conditions at the cascade exit.
    critical_cascade_input : dict
        Input conditions for critical state evaluation.
    fluid : object
        A fluid object with methods for thermodynamic property calculations.
    geometry : dict
        Dictionary containing geometric parameters of the cascade.
    angular_speed : float
        Angular speed of the cascade.
    results : dict
        Dictionary to store the evaluation results.
    model_options : dict
        Dictionary containing various model options.
    reference_values : dict
        Dictionary containing reference values for normalization.

    Returns
    -------
    residuals : dict
        A dictionary containing the residuals of the cascade evaluation.

    """

    # Define model options
    # TODO Add warnings that some settings have been assumed. This is a dangerous silent behavior as it is now
    # TODO It would make sense to define default values for these options at a higher level, for instance after processing the configuration file
    # TODO an advantage of the approach above is that it is possible to print a warning that will not be shown for each iteration of the solvers
    loss_model = model_options["loss_model"]
    deviation_model = model_options["deviation_model"]

    # Load reference values
    mass_flow_reference = reference_values["mass_flow_ref"]

    # Evaluate inlet plane
    inlet_plane = evaluate_cascade_inlet(
        cascade_inlet_input, fluid, geometry, angular_speed
    )

    # Evaluate exit plane
    cascade_exit_input["rothalpy"] = inlet_plane["rothalpy"]
    exit_plane, loss_dict = evaluate_cascade_exit(
        cascade_exit_input,
        fluid,
        geometry,
        inlet_plane,
        angular_speed,
        model_options["blockage_model"],
        loss_model,
    )

    # Evaluate isentropic enthalpy change
    props_is = fluid.get_props(cp.PSmass_INPUTS, exit_plane["p"], inlet_plane["s"])
    dh_is = exit_plane["h"] - props_is["h"]

    # Evaluate critical state
    residuals_critical, critical_state = cm.evaluate_choking(
        choking_input,
        inlet_plane,
        exit_plane,
        fluid,
        geometry,
        angular_speed,
        model_options,
        reference_values,
    )

    # Create dictionary with equation residuals
    mass_error_exit = inlet_plane["mass_flow"] - exit_plane["mass_flow"]
    residuals = {
        "loss_error_exit": exit_plane["loss_error"],
        "mass_error_exit": mass_error_exit / mass_flow_reference,
        **residuals_critical,
    }

    # Return plane data in dataframe
    results["plane"].loc[len(results["plane"]), inlet_plane.keys()] = inlet_plane
    results["plane"].loc[len(results["plane"]), exit_plane.keys()] = exit_plane

    # Return cascade data in dataframe
    cascade_data = {
        **loss_dict,
        **critical_state,
        "dh_s": dh_is,
        "incidence": inlet_plane["beta"] - geometry["leading_edge_angle"],
    }

    results["cascade"].loc[len(results["cascade"]), cascade_data.keys()] = cascade_data

    return residuals


def evaluate_cascade_inlet(cascade_inlet_input, fluid, geometry, angular_speed):
    r"""
    Evaluate the inlet plane of a cascade including velocity triangles,
    thermodynamic properties, and flow characteristics.

    This function calculates performance data at the inlet of a cascade based on the cascade geometry,
    fluid, and flow conditions. It computes velocity triangles, static and stagnation properties,
    Reynolds and Mach numbers, and the mass flow rate at the inlet.

    Parameters
    ----------
    cascade_inlet_input : dict
        Input parameters specific to the cascade inlet, including stagnation enthalpy (`h0`),
        entropy (`s`), absolute velocity (`v`), and flow angle (`alpha`).
    fluid : object
        A fluid object with methods for thermodynamic property calculations.
    geometry : dict
        Geometric parameters of the cascade such as radius at the mean inlet (`radius_mean_in`),
        chord length, and inlet area (`A_in`).
    angular_speed : float
        Angular speed of the cascade.

    Returns
    -------
    dict
        A dictionary of calculated parameters at the cascade inlet.
    """
    # Load cascade inlet input
    h0 = cascade_inlet_input["h0"]
    s = cascade_inlet_input["s"]
    v = cascade_inlet_input["v"]
    alpha = cascade_inlet_input["alpha"]

    # Load geometry
    radius = geometry["radius_mean_in"]
    chord = geometry["chord"]
    area = geometry["A_in"]

    # Calculate velocity triangles
    blade_speed = radius * angular_speed
    velocity_triangle = evaluate_velocity_triangle_in(blade_speed, v, alpha)
    w = velocity_triangle["w"]
    w_m = velocity_triangle["w_m"]

    # Calculate static properties
    h = h0 - 0.5 * v**2
    static_properties = fluid.get_props(cp.HmassSmass_INPUTS, h, s)
    rho = static_properties["d"]
    mu = static_properties["mu"]
    a = static_properties["a"]

    # Calculate stagnation properties
    stagnation_properties = fluid.get_props(cp.HmassSmass_INPUTS, h0, s)
    stagnation_properties = utils.add_string_to_keys(stagnation_properties, "0")

    # Calculate relative stagnation properties
    h0_rel = h + 0.5 * w**2
    relative_stagnation_properties = fluid.get_props(cp.HmassSmass_INPUTS, h0_rel, s)
    relative_stagnation_properties = utils.add_string_to_keys(
        relative_stagnation_properties, "0_rel"
    )

    # Calculate mach, reynolds and mass flow rate for cascade inlet
    Ma = v / a
    Ma_rel = w / a
    Re = rho * w * chord / mu
    m = rho * w_m * area
    rothalpy = h0_rel - 0.5 * blade_speed**2

    # Store results in dictionary
    plane = {
        **velocity_triangle,
        **static_properties,
        **stagnation_properties,
        **relative_stagnation_properties,
        "Ma": Ma,
        "Ma_rel": Ma_rel,
        "Re": Re,
        "mass_flow": m,
        "rothalpy": rothalpy,
    }

    return plane


def evaluate_cascade_exit(
    cascade_exit_input,
    fluid,
    geometry,
    inlet_plane,
    angular_speed,
    blockage,
    loss_model,
):
    r"""

    Evaluate the exit plane of a cascade including velocity triangles,
    thermodynamic properties, flow characteristics and loss coefficients.

    This function calculates performance data at the exit of a cascade based on the cascade geometry,
    fluid, and flow conditions. It computes velocity triangles, static and stagnation properties,
    Reynolds and Mach numbers, the mass flow rate and loss coefficients at the exit.

    Parameters
    ----------
    cascade_exit_input : dict
        Input parameters specific to the cascade exit, including relative velocity (`w`),
        relative flow angle (`beta`), entropy (`s`), and rothalpy.
    fluid : object
        A fluid object with methods for thermodynamic property calculations.
    geometry : dict
        Geometric parameters of the cascade such as chord length, opening, and area.
    inlet_plane : dict
        performance data at the inlet plane of the cascade.
    angular_speed : float
        Angular speed of the cascade.
    blockage : str or float or None
        The method or value for determining the throat blockage. It can be
        a string specifying a model name (`flat_plate_turbulent`), a numeric
        value between 0 and 1 representing the blockage factor directly, or
        None to use a default calculation method.
    loss_model : str
        The loss model used for calculating loss coefficients.

    Returns
    -------
    dict
        A dictionary of calculated parameters at the cascade exit.
    dict
        A dictionary of loss coefficients as calculated by the loss model.

    """

    # Load cascade exit variables
    w = cascade_exit_input["w"]
    beta = cascade_exit_input["beta"]
    s = cascade_exit_input["s"]
    rothalpy = cascade_exit_input["rothalpy"]

    # Load geometry
    chord = geometry["chord"]
    opening = geometry["opening"]
    area = geometry["A_out"]
    radius = geometry["radius_mean_out"]

    # Calculate velocity triangles
    blade_speed = angular_speed * radius
    velocity_triangle = evaluate_velocity_triangle_out(blade_speed, w, beta)
    v = velocity_triangle["v"]
    w_m = velocity_triangle["w_m"]

    # Calculate static properties
    h = rothalpy + 0.5 * blade_speed**2 - 0.5 * w**2
    static_properties = fluid.get_props(cp.HmassSmass_INPUTS, h, s)
    rho = static_properties["d"]
    mu = static_properties["mu"]
    a = static_properties["a"]

    # Calculate stagnation properties
    h0 = h + 0.5 * v**2
    stagnation_properties = fluid.get_props(cp.HmassSmass_INPUTS, h0, s)
    stagnation_properties = utils.add_string_to_keys(stagnation_properties, "0")

    # Calculate relative stagnation properties
    h0_rel = h + 0.5 * w**2
    relative_stagnation_properties = fluid.get_props(cp.HmassSmass_INPUTS, h0_rel, s)
    relative_stagnation_properties = utils.add_string_to_keys(
        relative_stagnation_properties, "0_rel"
    )

    # Calculate mach, reynolds and mass flow rate for cascade inlet
    Ma = v / a
    Ma_rel = w / a
    Re = rho * w * chord / mu
    rothalpy = h0_rel - 0.5 * blade_speed**2

    # Compute mass flow rate
    blockage_factor = compute_blockage_boundary_layer(blockage, Re, chord, opening)
    mass_flow = rho * w_m * area * (1 - blockage_factor)

    # Evaluate loss coefficient
    min_val = 1e-3
    loss_model_input = {
        "geometry": geometry,
        "loss_model": loss_model,
        "flow": {
            "p0_rel_in": inlet_plane["p0_rel"],
            "p0_rel_out": relative_stagnation_properties["p0_rel"],
            "p_in": inlet_plane["p"],
            "p_out": static_properties["p"],
            "beta_out": beta,
            "beta_in": inlet_plane["beta"],
            "Ma_rel_in": max(min_val, inlet_plane["Ma_rel"]),
            "Ma_rel_out": max(min_val, Ma_rel),
            "Re_in": max(min_val, inlet_plane["Re"]),
            "Re_out": max(min_val, Re),
            "gamma_out": static_properties["gamma"],
        },
    }

    # Compute loss coefficient from loss model
    loss_dict = lm.evaluate_loss_model(loss_model, loss_model_input)

    # Store results in dictionary
    plane = {
        **velocity_triangle,
        **static_properties,
        **stagnation_properties,
        **relative_stagnation_properties,
        **loss_dict,
        "Ma": Ma,
        "Ma_rel": Ma_rel,
        "Re": Re,
        "mass_flow": mass_flow,
        "rothalpy": rothalpy,
        "blockage": blockage_factor,
    }
    return plane, loss_dict


def evaluate_cascade_throat(
    cascade_throat_input,
    fluid,
    geometry,
    inlet_plane,
    angular_speed,
    blockage,
    loss_model,
):
    r"""
    Evaluate the throat plane of a cascade including velocity triangles,
    thermodynamic properties, flow characteristics and loss coefficients.

    This function calculates performance data at the throat of a cascade based on the cascade geometry,
    fluid, and flow conditions. It computes velocity triangles, static and stagnation properties,
    Reynolds and Mach numbers, the mass flow rate and loss coefficients at the throat.

    Parameters
    ----------
    cascade_throat_input : dict
        Input parameters specific to the cascade throat, including relative velocity (`w`),
        relative flow angle (`beta`), entropy (`s`), and rothalpy.
    fluid : object
        A fluid object with methods for thermodynamic property calculations.
    geometry : dict
        Geometric parameters of the cascade such as chord length, opening, and area.
    inlet_plane : dict
        performance data at the inlet plane of the cascade.
    angular_speed : float
        Angular speed of the cascade.
    blockage : str or float or None
        The method or value for determining the throat blockage. It can be
        a string specifying a model name (`flat_plate_turbulent`), a numeric
        value between 0 and 1 representing the blockage factor directly, or
        None to use a default calculation method.
    loss_model : str
        The loss model used for calculating loss coefficients.

    Returns
    -------
    dict
        A dictionary of calculated parameters at the cascade throat.
    dict
        A dictionary of loss coefficients as calculated by the loss model.

    """

    # Load cascade exit variables
    w = cascade_throat_input["w"]
    beta = cascade_throat_input["beta"]
    s = cascade_throat_input["s"]
    rothalpy = cascade_throat_input["rothalpy"]

    # Load geometry
    chord = geometry["chord"]
    opening = geometry["opening"]
    area = geometry["A_throat"]
    radius = geometry["radius_mean_throat"]

    # Calculate velocity triangles
    blade_speed = angular_speed * radius
    velocity_triangle = evaluate_velocity_triangle_out(blade_speed, w, beta)
    v = velocity_triangle["v"]

    # Calculate static properties
    h = rothalpy + 0.5 * blade_speed**2 - 0.5 * w**2
    static_properties = fluid.get_props(cp.HmassSmass_INPUTS, h, s)
    rho = static_properties["d"]
    mu = static_properties["mu"]
    a = static_properties["a"]

    # Calculate stagnation properties
    h0 = h + 0.5 * v**2
    stagnation_properties = fluid.get_props(cp.HmassSmass_INPUTS, h0, s)
    stagnation_properties = utils.add_string_to_keys(stagnation_properties, "0")

    # Calculate relative stagnation properties
    h0_rel = h + 0.5 * w**2
    relative_stagnation_properties = fluid.get_props(cp.HmassSmass_INPUTS, h0_rel, s)
    relative_stagnation_properties = utils.add_string_to_keys(
        relative_stagnation_properties, "0_rel"
    )

    # Calculate mach, reynolds and mass flow rate for cascade inlet
    Ma = v / a
    Ma_rel = w / a
    Re = rho * w * chord / mu
    rothalpy = h0_rel - 0.5 * blade_speed**2

    # Compute mass flow rate
    blockage_factor = compute_blockage_boundary_layer(blockage, Re, chord, opening)
    mass_flow = rho * w * area * (1 - blockage_factor)

    # Evaluate loss coefficient
    min_val = 1e-3
    loss_model_input = {
        "geometry": geometry,
        "loss_model": loss_model,
        "flow": {
            "p0_rel_in": inlet_plane["p0_rel"],
            "p0_rel_out": relative_stagnation_properties["p0_rel"],
            "p_in": inlet_plane["p"],
            "p_out": static_properties["p"],
            "beta_out": beta,
            "beta_in": inlet_plane["beta"],
            "Ma_rel_in": max(min_val, inlet_plane["Ma_rel"]),
            "Ma_rel_out": max(min_val, Ma_rel),
            "Re_in": max(min_val, inlet_plane["Re"]),
            "Re_out": max(min_val, Re),
            "gamma_out": static_properties["gamma"],
        },
    }

    # Compute loss coefficient from loss model
    loss_dict = lm.evaluate_loss_model(loss_model, loss_model_input)

    # Store results in dictionary
    plane = {
        **velocity_triangle,
        **static_properties,
        **stagnation_properties,
        **relative_stagnation_properties,
        **loss_dict,
        "Ma": Ma,
        "Ma_rel": Ma_rel,
        "Re": Re,
        "mass_flow": mass_flow,
        "rothalpy": rothalpy,
        "blockage": blockage_factor,
    }
    return plane, loss_dict


def evaluate_cascade_interspace(
    h0_exit,
    v_m_exit,
    v_t_exit,
    rho_exit,
    radius_exit,
    area_exit,
    blockage_exit,
    radius_inlet,
    area_inlet,
    fluid,
):
    """
    Calculate the inlet conditions for the next cascade based on the exit conditions of the previous cascade.

    This function computes the inlet thermodynamic and velocity conditions for the next cascade using the exit conditions
    from the previous cascade and the flow equations for the interspace between cascades.

    Assumptions:

    - Conservation of stagnation enthalpy
    - Conservation of angular momentum
    - Negligible change in density


    Parameters
    ----------
    h0_exit : float
        Stagnation enthalpy at the exit of the previous cascade.
    v_m_exit : float
        Meridional component of velocity at the exit of the previous cascade.
    v_t_exit : float
        Tangential component of velocity at the exit of the previous cascade.
    rho_exit : float
        Fluid density at the exit of the previous cascade.
    radius_exit : float
        Radius at the exit of the previous cascade.
    area_exit : float
        Flow area at the exit of the previous cascade.
    radius_inlet : float
        Radius at the inlet of the next cascade.
    area_inlet : float
        Flow area at the inlet of the next cascade.
    fluid : object
        A fluid object with methods for thermodynamic property calculations.

    Returns
    -------
    float
        Stagnation enthalpy at the inlet of the next cascade.
    float
        Entropy at the inlet of the next cascade.
    float
        Flow angle at the inlet of the next cascade (in degrees).
    float
        Velocity at the inlet of the next cascade.

    Warnings
    --------
    The assumption of constant density leads to a small inconsistency in the thermodynamic state,
    manifesting as a slight variation in entropy across the interspace. In future versions of the
    code, it is recommended to evaluate the interspace with a 1D model for the flow in annular ducts
    to improve accuracy and consistency in the analysis.
    """

    # Assume no heat transfer
    h0_in = h0_exit

    # Assume no friction (angular momentum is conserved)
    v_t_in = v_t_exit * radius_exit / radius_inlet

    # Assume density variation is negligible
    v_m_in = v_m_exit * area_exit / area_inlet * (1 - blockage_exit)

    # Compute velocity vector
    v_in = np.sqrt(v_t_in**2 + v_m_in**2)
    alpha_in = math.arctand(v_t_in / v_m_in)

    # Compute thermodynamic state
    h_in = h0_in - 0.5 * v_in**2
    rho_in = rho_exit
    stagnation_properties = fluid.get_props(cp.DmassHmass_INPUTS, rho_in, h_in)
    s_in = stagnation_properties["s"]

    return h0_in, s_in, alpha_in, v_in


def evaluate_velocity_triangle_in(blade_speed, v, alpha):
    """
    Compute the velocity triangle at the inlet of the cascade.

    This function calculates the components of the velocity triangle at the
    inlet of a cascade, based on the blade speed, absolute velocity, and
    absolute flow angle. It computes both the absolute and relative velocity
    components in the meridional and tangential directions, as well as the
    relative flow angle.

    Parameters
    ----------
    blade_speed : float
        Blade speed.
    v : float
        Absolute velocity.
    alpha : float
        Absolute flow angle in radians.

    Returns
    -------
    dict
        A dictionary containing the following properties:

        - `blade_speed` (float): Blade velocity.
        - `v` (float): Absolute velocity.
        - `v_m` (float): Meridional component of absolute velocity.
        - `v_t` (float): Tangential component of absolute velocity.
        - `alpha` (float): Absolute flow angle in radians.
        - `w` (float): Relative velocity magnitude.
        - `w_m` (float): Meridional component of relative velocity.
        - `w_t` (float): Tangential component of relative velocity.
        - `beta` (float): Relative flow angle in radians.
    """

    # Absolute velocities
    v_t = v * math.sind(alpha)
    v_m = v * math.cosd(alpha)

    # Relative velocities
    w_t = v_t - blade_speed
    w_m = v_m
    w = np.sqrt(w_t**2 + w_m**2)

    # Relative flow angle
    beta = math.arctand(w_t / w_m)

    # Store in dict
    vel_in = {
        "blade_speed": blade_speed,
        "v": v,
        "v_m": v_m,
        "v_t": v_t,
        "alpha": alpha,
        "w": w,
        "w_m": w_m,
        "w_t": w_t,
        "beta": beta,
    }

    return vel_in


def evaluate_velocity_triangle_out(blade_speed, w, beta):
    """
    Compute the velocity triangle at the outlet of the cascade.

    This function calculates the components of the velocity triangle at the
    outlet of a cascade, based on the blade speed, relative velocity, and
    relative flow angle. It computes both the absolute and relative velocity
    components in the meridional and tangential directions, as well as the
    absolute flow angle.

    Parameters
    ----------
    blade_speed : float
        Blade speed.
    w : float
        Relative velocity.
    beta : float
        Relative flow angle in radians.

    Returns
    -------
    dict
        A dictionary containing the following properties:

        - `blade_speed` (float): Blade velocity.
        - `v` (float): Absolute velocity.
        - `v_m` (float): Meridional component of absolute velocity.
        - `v_t` (float): Tangential component of absolute velocity.
        - `alpha` (float): Absolute flow angle in radians.
        - `w` (float): Relative velocity magnitude.
        - `w_m` (float): Meridional component of relative velocity.
        - `w_t` (float): Tangential component of relative velocity.
        - `beta` (float): Relative flow angle in radians.
    """

    # Relative velocities
    w_t = w * math.sind(beta)
    w_m = w * math.cosd(beta)

    # Absolute velocities
    v_t = w_t + blade_speed
    v_m = w_m
    v = np.sqrt(v_t**2 + v_m**2)

    # Absolute flow angle
    alpha = math.arctand(v_t / v_m)

    # Store in dict
    vel_out = {
        "blade_speed": blade_speed,
        "v": v,
        "v_m": v_m,
        "v_t": v_t,
        "alpha": alpha,
        "w": w,
        "w_m": w_m,
        "w_t": w_t,
        "beta": beta,
    }

    return vel_out


def compute_blockage_boundary_layer(blockage_model, Re, chord, opening):
    r"""
    Calculate the blockage factor due to boundary layer displacement thickness.

    This function computes the blockage factor caused by the boundary layer
    displacement thickness at a given flow station. The blockage factor affects
    mass flow rate calculations effectively reducing the flow area.

    The blockage factor can be determined through various methods, including:

        1. Calculation based on a correlation for the displacement thickness of turbulent boundary layer over a flat plate with zero pressure gradient.
        2. Using a numerical value specified directly by the user.

    The correlation for turbulent boundary layer displacement thickness over a flat plate is given by :cite:`cengel_fluid_2014`:

    .. math::
        \delta^* = \frac{0.048}{Re^{1/5}} \times 0.9 \times \text{chord}

    From this the blockage factor is calculated as

    .. math::
        \text{blockage_factor} = 2 \times \frac{\delta^*}{\text{opening}}

    Parameters
    ----------
    blockage_model : str or float or None
        The method or value for determining the blockage factor. It can be
        a string specifying a model name (`flat_plate_turbulent`), a numeric
        value between 0 and 1 representing the blockage factor directly, or
        None to use a default calculation method.
    Re : float
        Reynolds number, used if `blockage_model` is `flat_plate_turbulent`.
    chord : float
        Chord length, used if `blockage_model` is `flat_plate_turbulent`.
    opening : float
        Throat opening, used if `blockage_model` is `flat_plate_turbulent`.

    Returns
    -------
    float
        The calculated blockage factor, a value between 0 and 1, where 1
        indicates full blockage and 0 indicates no blockage.

    Raises
    ------
    ValueError
        If `blockage_model` is an invalid option, or required parameters
        for the chosen method are missing.
    """

    if blockage_model == BLOCKAGE_MODELS[0]:
        displacement_thickness = 0.048 / Re ** (1 / 5) * 0.9 * chord
        blockage_factor = 2 * displacement_thickness / opening

    elif isinstance(blockage_model, (float, int)) and 0 <= blockage_model <= 1:
        blockage_factor = blockage_model

    elif blockage_model is None:
        blockage_factor = 0.00

    else:
        raise ValueError(
            f"Invalid throat blockage option: '{blockage_model}'. "
            "Valid options are 'flat_plate_turbulent', a numeric value between 0 and 1, or None."
        )

    return blockage_factor


def compute_efficiency_breakdown(results):
    """
    Compute the loss of total-to-static efficiency due to each various loss component in cascades.

    This function calculates the fraction of total-to-static efficiency drop attributed to each loss component
    in a turbine cascade. A correction factor for the re-heating effect is applied to align the sum of individual
    cascade losses with the total observed change in enthalpy and ensure consistency with
    the overall total-to-static efficiency.

    Parameters
    ----------
    results : dict
        The data for the cascades, including plane and cascade specific parameters.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing efficiency drop fractions for each loss type in each cascade.
        Columns are named as `efficiency_drop_{loss_type}`, where `loss_type` includes:

        - `profile`
        - `incidence`
        - `secondary`
        - `clearance`
        - `trailing`

    """
    # Load parameters
    h_out_s = results["reference_values"]["h_out_s"]
    h0_in = results["plane"]["h0"].values[0]
    h_out = results["plane"]["h"].values[-1]
    cascade = results["cascade"]
    number_of_cascades = results["geometry"]["number_of_cascades"]

    # Compute a correction factor due to re-heating effect
    dhs_total = h_out - h_out_s
    dhs_sum = cascade["dh_s"].sum()
    correction = dhs_total / dhs_sum

    # Initialize DataFrame
    loss_types = ["profile", "incidence", "secondary", "clearance", "trailing"]
    breakdown = pd.DataFrame(columns=[f"efficiency_drop_{type}" for type in loss_types])

    # Loss breakdown in each cascade
    for i in range(number_of_cascades):
        # Construct column names dynamically based on loss_types
        col_names = [f"loss_{type}" for type in loss_types]

        # Retrieve values for each specified column and compute the fractions
        fracs = cascade.loc[i, col_names] / cascade.loc[i, "loss_total"]
        dh_s = cascade["dh_s"][i]
        efficiency_drop = correction * dh_s / (h0_in - h_out_s)

        # Append efficiency drop breakdown
        breakdown.loc[len(breakdown)] = (fracs * efficiency_drop).values

    return breakdown


def compute_stage_performance(results):
    r"""
    Calculate the stage performance metrics of the turbine.

    This function extracts necessary values from the `results` dictionary, performs calculations to determine
    stage performance metrics, and returns these in a dictionary.

    Parameters
    ----------
    results : dict
        A dictionary containing all necessary information, such as geometry and flow characteristics.

    Returns
    -------
    dict
        A dictionary with calculated stage parameters.

    """

    # Only proceed if there are stages
    number_of_stages = results["geometry"]["number_of_stages"]
    if number_of_stages == 0:
        return {}

    # Calculate the degree of reaction for each stage using list comprehension
    h = results["plane"]["h"].values
    R = np.array(
        [
            (h[i * 4 + 1] - h[i * 4 + 3]) / (h[i * 4] - h[i * 4 + 3])
            for i in range(number_of_stages)
        ]
    )

    # Store all variables in dictionary
    stages = {"reaction": R}

    return stages


def compute_overall_performance(results, geometry):
    """
    Calculate the overall performance metrics of the turbine.

    This function extracts necessary values from the `results` dictionary, performs calculations to determine
    overall performance metrics, and returns these in a dictionary.

    Parameters
    ----------
    results : dict
        A dictionary containing all necessary information, such as geometry and flow characteristics.

    Returns
    -------
    dict
        An dictionary containing the calculated performance metrics.
    """

    angular_speed = results["boundary_conditions"]["omega"]
    v0 = results["reference_values"]["v0"]
    h_out_s = results["reference_values"]["h_out_s"]
    d_out_s = results["reference_values"]["d_out_s"]

    # Calculation of overall performance
    p = results["plane"]["p"].values
    p0 = results["plane"]["p0"].values
    h0 = results["plane"]["h0"].values
    v_out = results["plane"]["v"].values[-1]
    u_out = results["plane"]["blade_speed"].values[-1]
    mass_flow = results["plane"]["mass_flow"].values[-1]
    exit_flow_angle = results["plane"]["alpha"].values[-1]
    PR_tt = p0[0] / p0[-1]
    PR_ts = p0[0] / p[-1]
    h0_in = h0[0]
    h0_out = h0[-1]
    efficiency_tt = (h0_in - h0_out) / (h0_in - h_out_s - 0.5 * v_out**2) * 100
    efficiency_ts = (h0_in - h0_out) / (h0_in - h_out_s) * 100
    efficiency_ts_drop_kinetic = 0.5 * v_out**2 / (h0_in - h_out_s)
    efficiency_ts_drop_losses = 1.0 - efficiency_ts - efficiency_ts_drop_kinetic
    power = mass_flow * (h0_in - h0_out)
    torque = power / angular_speed
    specific_speed = (
        angular_speed * (mass_flow / d_out_s) ** 0.5 / ((h0_in - h_out_s) ** 0.75)
    )

    # Store all variables in dictionary
    overall = {
        "PR_tt": PR_tt,
        "PR_ts": PR_ts,
        "mass_flow_rate": mass_flow,
        "efficiency_tt": efficiency_tt,
        "efficiency_ts": efficiency_ts,
        "efficiency_ts_drop_kinetic": efficiency_ts_drop_kinetic,
        "efficiency_ts_drop_losses": efficiency_ts_drop_losses,
        "power": power,
        "torque": torque,
        "angular_speed": angular_speed,
        "exit_flow_angle": exit_flow_angle,
        "exit_velocity": v_out,
        "spouting_velocity": v0,
        "last_blade_velocity": u_out,
        "blade_jet_ratio": u_out / v0,
        "h0_in": h0_in,
        "h0_out": h0_out,
        "h_out_s": h_out_s,
        "specific_speed": specific_speed,
        "blade_jet_ratio_hub": angular_speed * geometry["radius_hub_out"][-1] / v0,
        "blade_jet_ratio_mean": angular_speed * geometry["radius_mean_out"][-1] / v0,
        "blade_jet_ratio_tip": angular_speed * geometry["radius_tip_out"][-1] / v0,
    }

    return overall



import os
import time
import yaml
import copy
import datetime
import threading
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from ..config_validation import read_configuration_file
from .. import utilities
from .. import pysolver_view as psv
from .. import fluid_properties as props

from . import brayton_recuperated
from . import brayton_split_compression


utilities.set_plot_options()
COLORS_MATLAB = utilities.COLORS_MATLAB
LABEL_MAPPING = {
    "s": "Entropy [J/kg/K]",
    "T": "Temperature [K]",
    "p": "Pressure [Pa]",
    "h": "Enthalpy [J/kg]",
    "d": "Density [kg/m$^3$]",
    "a": "Speed of sound [m/s]",
    "Z": "Compressibility factor [-]",
    "heat": "Heat flow rate [W]",
}

# Cycle configurations available
CYCLE_TOPOLOGIES = {
    "recuperated": brayton_recuperated.evaluate_cycle,
    "split_compression": brayton_split_compression.evaluate_cycle,
    "recompression": brayton_split_compression.evaluate_cycle,
}

GRAPHICS_PLACEHOLDER = {
    "process_lines": {},
    "state_points": {},
    "pinch_point_lines": {},
}


class ThermodynamicCycleOptimization:
    def __init__(self, config_file, out_dir=None):
        """
        Initializes the optimization manager with a configuration file.
        Parameters:
            config_file (str): The path to the YAML configuration file.
        """

        # Define filename with unique date-time identifier
        if out_dir == None:
            current_time = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
            self.out_dir = f"results/case_{current_time}"

        # Create a directory to save simulation results
        if not os.path.exists(self.out_dir):
            os.makedirs(self.out_dir)

        self.config = self.load_config(config_file)
        self.problem = self.setup_problem()
        self.solver = self.setup_solver()

    def load_config(self, config_file):
        """
        Loads configuration from a YAML file.
        """
        return read_configuration_file(config_file)

    def setup_problem(self):
        """
        Sets up the ThermodynamicCycleProblem based on the loaded configuration.
        """
        self.problem = ThermodynamicCycleProblem(self.config["problem_formulation"])
        self.problem.fitness(self.problem.x0)
        return self.problem

    def setup_solver(self):
        """
        Configures and returns the optimization solver.
        """

        # Optimize the thermodynamic cycle
        self.solver = psv.OptimizationSolver(
            self.problem,
            **self.config["solver_options"],
            callback_functions=[
                self.problem.plot_cycle_callback,
                self.problem.save_config_callback,
            ],
        )

        return self.solver

    def run_optimization(self):
        """
        Executes the optimization process.
        """
        self.solver.solve(self.problem.x0)

    def save_results(self):
        """
        Saves the results of the optimization, including configurations and output files.
        """
        filename = os.path.join(self.out_dir, "optimal_solution")
        self.problem.to_excel(filename=filename + ".xlsx")
        self.problem.save_current_configuration(filename=filename + ".yaml")

        # Print final solution values
        print()
        print("Optimal set of design variables (normalized)")
        for key, value in self.problem.vars_normalized.items():
            print(f"{key:40}: {value:0.3f}")

    def generate_output_files(self):
        """
        Generates additional output files, such as Excel files or plots.
        """
        # Code to handle output generation like creating plots or exporting to Excel

    def create_animation(self):
        """
        Creates an animation of the optimization history.
        """
        # Code to generate animation from the optimization results


class ThermodynamicCycleProblem(psv.OptimizationProblem):
    """
    A class to represent a thermodynamic cycle optimization problem.

    This class provides functionalities to load and update the configuration for
    the thermodynamic cycle, and to perform interactive plotting based on the current
    configuration.

    Attributes
    ----------
    plot_initialized : bool
        Flag to indicate if the plot has been initialized.
    constraints : dict
        Dictionary holding the constraints of the problem.
    fixed_parameters : dict
        Dictionary holding the fixed parameters of the problem.
    design_variables : dict
        Dictionary holding the current values of the design variables.
    lower_bounds : dict
        Dictionary holding the lower bounds of the design variables.
    upper_bounds : dict
        Dictionary holding the upper bounds of the design variables.
    keys : list
        List of keys (names) of the design variables.
    x0 : np.ndarray
        Initial guess for the optimization problem.

    Methods
    -------
    update_config(configuration):
        Update the problem's configuration based on the provided dictionary.
    load_config_from_file(configuration_file):
        Load and update the problem's configuration from a specified file.
    plot_cycle_interactive(configuration_file, update_interval=0.20):
        Perform interactive plotting, updating the plot based on the configuration file.
    """

    def __init__(self, configuration, out_dir=None):
        """
        Constructs all the necessary attributes for the ThermodynamicCycleProblem object.

        Parameters
        ----------
        config : dict
            Dictionary containing the configuration for the thermodynamic cycle problem.
        """
        # Initialize variables
        self.figure = None
        self.figure_TQ = None
        self.configuration = copy.deepcopy(configuration)
        self.update_configuration(self.configuration)
        self.graphics = copy.deepcopy(GRAPHICS_PLACEHOLDER)

        # Create fluid object to calculate states used for the scaling of design variables
        self.fluid = props.Fluid(**self.fixed_parameters["working_fluid"])
        self._calculate_special_points()

        # Define filename with unique date-time identifier
        if out_dir == None:
            current_time = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
            self.out_dir = f"results/case_{current_time}"

        # Create a directory to save simulation results
        if not os.path.exists(self.out_dir):
            os.makedirs(self.out_dir)

        # # Save the initial configuration as YAML file
        # self.save_current_configuration("initial_configuration.yaml")

    def update_configuration(self, configuration):
        """
        Update the problem's configuration based on the provided dictionary.

        Parameters
        ----------
        config : dict
            Dictionary containing the new configuration for the thermodynamic cycle problem.
        """
        conf = configuration
        self.cycle_topology = conf["cycle_topology"]
        self.plot_settings = conf["plot_settings"]
        self.constraints = conf["constraints"]
        self.fixed_parameters = conf["fixed_parameters"]
        self.objective_function = conf["objective_function"]
        self.vars_normalized = {
            k: v["value"] for k, v in conf["design_variables"].items()
        }
        self.lower_bounds = {k: v["min"] for k, v in conf["design_variables"].items()}
        self.upper_bounds = {k: v["max"] for k, v in conf["design_variables"].items()}
        self.keys = list(self.vars_normalized.keys())
        self.x0 = np.asarray([var for var in self.vars_normalized.values()])

    def _calculate_special_points(self):
        """
        Calculates and stores key thermodynamic states for the fluid used in the cycle.

        This method computes two specific thermodynamic states:
        1. Saturated liquid state at the heat sink inlet temperature.
        2. Dilute gas state at the heat source inlet temperature.

        If the heat sink inlet temperature is below the fluid's critical temperature,
        the saturated liquid state is calculated at this temperature. If it is above
        the critical temperature, the state is calculated along the pseudocritical line.

        The dilute gas state is calculated at the heat source inlet temperature and at
        very low pressure (slightly above the triple pressure)

        The calculated states, along with the fluid's critical and triple point properties,
        are stored in the `fixed_parameters` dictionary under the 'fluid' key.

        These states are intended to define the bounds of the design variables specified
        in the YAML configuration file, for example:

            compressor_inlet_enthalpy:
            min: 0.9*$fluid.liquid_sink_temperature.h
            max: 2.0*$fluid.liquid_sink_temperature.h
            value: 0.2
            turbine_inlet_pressure:
            min: 0.75*$fluid.critical_point.p
            max: 5.00*$fluid.critical_point.p
            value: 0.5
            turbine_inlet_enthalpy:
            min: 1.10*$fluid.critical_point.h
            max: 1.00*$fluid.gas_source_temperature.h
            value: 0.90

        """
        # Compute saturated liquid state at sink temperature
        # Use pseudocritical line if heat sink is above critical temperature
        T_sink = self.fixed_parameters["heat_sink"]["inlet_temperature"]
        crit = self.fluid.critical_point
        if T_sink < crit.T:
            state_sat = self.fluid.set_state(props.QT_INPUTS, 0.0, T_sink)
        else:
            state_sat = self.fluid.set_state(props.DmassT_INPUTS, crit.rho, T_sink)

        # Compute dilute gas state at heat source temperature
        T_source = self.fixed_parameters["heat_source"]["inlet_temperature"]
        p_triple = 1.01 * self.fluid.triple_point_liquid.p
        state_dilute = self.fluid.set_state(props.PT_INPUTS, p_triple, T_source)

        # Save states in the fixed parameters dictionary
        self.fixed_parameters_bis = copy.deepcopy(self.fixed_parameters)
        self.fixed_parameters_bis["working_fluid"] = {
            "critical_point": self.fluid.critical_point.to_dict(),
            "triple_point_liquid": self.fluid.triple_point_liquid.to_dict(),
            "triple_point_vapor": self.fluid.triple_point_vapor.to_dict(),
            "liquid_at_heat_sink_temperature": state_sat.to_dict(),
            "gas_at_heat_source_temperature": state_dilute.to_dict(),
        }

    def load_configuration_file(self, config_file):
        """
        Load and update the problem's configuration from a specified file.

        Useful to plot the cycle according to the latest version of the configuration
        file in real time (interactive initial guess generation)

        Parameters
        ----------
        config_file : str
            Path to the configuration file.
        """
        config = utilities.read_configuration_file(config_file)
        self.update_configuration(config["problem_formulation"])
        self._calculate_special_points()

    def save_current_configuration(self, filename):
        """Save the current configuration to a YAML file."""
        config_data = {k: v for k, v in self.configuration.items()}
        config_data = utilities.convert_numpy_to_python(config_data, precision=12)
        with open(filename, "w") as file:
            yaml.dump(config_data, file, default_flow_style=False, sort_keys=False)

    def save_config_callback(self, x, iter):
        """
        A callback function to save the current configuration during optimization iterations.

        Parameters:
        - x : The current solution vector from the optimizer.
        - iter : The current optimization iteration count.

        This function acts as a bridge between the optimizer callback requirements and the
        existing `save_current_configuration` function.
        """

        # Create a 'results' directory if it doesn't exist
        self.optimization_dir = os.path.join(self.out_dir, "optimization")
        os.makedirs(self.optimization_dir, exist_ok=True)

        # Define the filename using the solver's iteration number
        filename = os.path.join(self.optimization_dir, f"iteration_{iter:03d}.yaml")

        # Call the existing function to save the configuration
        self.save_current_configuration(filename)

    def fitness(self, x):
        """
        Evaluate optimization problem
        """

        # print("x", x)
        # Update configuration with the current values of x
        self.update_variables(x)

        # Evaluate thermodynamic cycle
        if self.cycle_topology in CYCLE_TOPOLOGIES.keys():
            self.cycle_data = CYCLE_TOPOLOGIES[self.cycle_topology](
                self.vars_physical,
                self.fixed_parameters,
                self.constraints,
                self.objective_function,
            )
        else:
            options = ", ".join(f"'{k}'" for k in CYCLE_TOPOLOGIES.keys())
            raise ValueError(
                f"Invalid cycle topology: '{self.cycle_topology}'. Available options: {options}"
            )

        # Define objective function and constraints
        self.f = self.cycle_data["objective_function"]
        self.c_eq = self.cycle_data["equality_constraints"]
        self.c_ineq = self.cycle_data["inequality_constraints"]

        # Combine objective function and constraints
        out = psv.combine_objective_and_constraints(self.f, self.c_eq, self.c_ineq)

        return out

    def update_variables(self, x):
        """Update the problem variables based on the new values provided by the optimizer."""
        self.vars_normalized = dict(zip(self.keys, x))
        self.vars_physical = self._scale_normalized_to_physical(self.vars_normalized)
        for k, v in self.vars_normalized.items():
            if k in self.configuration["design_variables"]:
                self.configuration["design_variables"][k]["value"] = v
            else:
                # Optionally handle the error or log a warning if the key does not exist
                print(f"Warning: {k} is not a recognized design variable.")

    def get_bounds(self):
        dim = len(self.vars_normalized)
        return ([0.0] * dim, [1.00] * dim)

    def get_nec(self):
        return psv.count_constraints(self.c_eq)

    def get_nic(self):
        return psv.count_constraints(self.c_ineq)

    def _scale_normalized_to_physical(self, vars_normalized):
        # Define helper function
        def scale_value(min_val, max_val, normalized_val):
            return min_val + (max_val - min_val) * normalized_val

        # Loop over normalized variables
        vars_physical = {}
        for key, value in vars_normalized.items():
            try:
                # Fetch the bounds expressions
                lower_expr = self.lower_bounds[key]
                upper_expr = self.upper_bounds[key]

                # Try to evaluate bounds as expressions if they are strings
                if isinstance(lower_expr, str):
                    lower = utilities.render_and_evaluate(
                        lower_expr, self.fixed_parameters_bis
                    )
                else:
                    lower = lower_expr
                if isinstance(upper_expr, str):
                    upper = utilities.render_and_evaluate(
                        upper_expr, self.fixed_parameters_bis
                    )
                else:
                    upper = upper_expr

                vars_physical[key] = scale_value(lower, upper, value)

            except (KeyError, ValueError) as e:
                raise ValueError(f"Error processing bounds for '{key}': {e}")

        return vars_physical

    def to_excel(self, filename="performance.xlsx"):
        """
        Exports the cycle performance data to Excel file
        """

        # Define variable map
        variable_map = {
            "fluid_name": {"name": "fluid_name", "unit": "-"},
            "T": {"name": "temperature", "unit": "K"},
            "p": {"name": "pressure", "unit": "Pa"},
            "rho": {"name": "density", "unit": "kg/m3"},
            "Q": {"name": "quality", "unit": "-"},
            "Z": {"name": "compressibility_factor", "unit": "-"},
            "u": {"name": "internal_energy", "unit": "J/kg"},
            "h": {"name": "enthalpy", "unit": "J/kg"},
            "s": {"name": "entropy", "unit": "J/kg/K"},
            "cp": {"name": "isobaric_heat_capacity", "unit": "J/kg/K"},
            "cv": {"name": "isochoric_heat_capacity", "unit": "J/kg/K"},
            "gamma": {"name": "heat_capacity_ratio", "unit": "-"},
            "a": {"name": "speed_of_sound", "unit": "m/s"},
            "mu": {"name": "dynamic_viscosity", "unit": "Pa*s"},
            "k": {"name": "thermal_conductivity", "unit": "W/m/K"},
            "superheating": {"name": "superheating_degree", "unit": "K"},
            "subcooling": {"name": "subcooling_degree", "unit": "K"},
        }

        # Initialize a list to hold all rows of the DataFrame
        data_rows = []

        # Prepare the headers and units rows
        headers = ["state"]
        units_row = ["units"]

        for key in variable_map:
            headers.append(variable_map[key]["name"])
            units_row.append(variable_map[key]["unit"])

        # Iterate over each component in the dictionary
        for component_name, component in self.cycle_data["components"].items():
            if component["type"] == "heat_exchanger":
                # Handle heat exchanger sides separately
                for side in ["hot_side", "cold_side"]:
                    # Append the data for state_in and state_out to the rows list
                    state_in = component[side]["state_in"].to_dict()
                    state_out = component[side]["state_out"].to_dict()
                    data_rows.append(
                        [f"{component_name}_{side}_in"]
                        + [state_in.get(key, None) for key in variable_map]
                    )
                    data_rows.append(
                        [f"{component_name}_{side}_out"]
                        + [state_out.get(key, None) for key in variable_map]
                    )
            else:
                # Handle non-heat exchanger components
                # Append the data for state_in and state_out to the rows list
                state_in = component["state_in"].to_dict()
                state_out = component["state_out"].to_dict()
                data_rows.append(
                    [f"{component_name}_in"]
                    + [state_in.get(key, None) for key in variable_map]
                )
                data_rows.append(
                    [f"{component_name}_out"]
                    + [state_out.get(key, None) for key in variable_map]
                )

        # Create a DataFrame with data rows
        df = pd.DataFrame(data_rows, columns=headers)

        # Insert the units row
        df.loc[-1] = units_row  # Adding a row
        df.index = df.index + 1  # Shifting index
        df = df.sort_index()  # Sorting by index

        # # Export to Excel
        # df.to_excel(
        #     os.path.join(self.out_dir, filename),
        #     index=False,
        #     header=True,
        #     sheet_name="cycle_states",
        # )

        # Prepare energy_analysis data
        df_2 = pd.DataFrame(
            list(self.cycle_data["energy_analysis"].items()),
            columns=["Parameter", "Value"],
        )

        # Export to Excel
        # filename = os.path.join(self.out_dir, filename)
        with pd.ExcelWriter(filename, engine="openpyxl") as writer:
            df.to_excel(writer, index=False, sheet_name="cycle_states")
            df_2.to_excel(writer, index=False, sheet_name="energy_analysis")

    def plot_cycle(self):
        """
        Plots or updates the thermodynamic cycle diagrams based on current settings,
        including the option to include a pinch point diagram.

        This function is capable of both creating new cycle diagrams and updating existing ones.
        It's particularly useful in dynamic scenarios such as during optimization steps,
        where the plot needs to be refreshed continually with new data.
        The method also supports real-time updates based on the latest configuration settings.

        The function first determines the number of subplots required based on the cycle diagrams
        specified in the 'plot_settings' attribute and whether a pinch point diagram is included.
        It then either initializes a new figure and axes or updates existing ones.
        Each thermodynamic diagram (phase diagram and cycle components) and the optional pinch point
        diagram are plotted or updated accordingly.

        The method ensures that the plot reflects the current state of the cycle, including any
        changes during optimization or adjustments to configuration settings.

        The data of the plots can be updated interactively, but the subplot objects created can
        only be specified upon class initialization. For example, it is not possible to switch
        on and off the pinch point diagram or to add new thermodynamic diagrams. The class would
        have to be re-initialized upon those scenarios. The reason for this choice is that re-creating
        a figure would be too time consuming and not practical for the real-time updating of the plots

        """

        # Determine the number of subplots
        include_pinch_diagram = self.plot_settings.get("pinch_point_diagram", False)
        ncols = len(self.plot_settings["diagrams"]) + int(include_pinch_diagram)
        nrows = 1

        # Initialize the figure and axes
        if not (self.figure and plt.fignum_exists(self.figure.number)):
            # Reset the graphics objects if the figure was closed
            self.graphics = copy.deepcopy(GRAPHICS_PLACEHOLDER)
            self.figure, self.axes = plt.subplots(
                nrows, ncols, figsize=(5.2 * ncols, 4.8)
            )
            self.axes = utilities.ensure_iterable(self.axes)

        # Plot or update each thermodynamic_diagram
        for i, ax in enumerate(self.axes[:-1] if include_pinch_diagram else self.axes):
            plot_config = self.plot_settings["diagrams"][i]
            plot_config = self._get_diagram_default_settings(plot_config)
            self._plot_thermodynamic_diagram(ax, plot_config, i)

        # Plot or update the pinch point diagram
        if include_pinch_diagram:
            self._plot_pinch_point_diagram(self.axes[-1], ncols - 1)

        # Adjust layout and refresh plot
        self.figure.tight_layout(pad=2)
        plt.draw()
        plt.pause(0.05)

    def plot_cycle_realtime(self, configuration_file, update_interval=0.1):
        """
        Perform interactive plotting, updating the plot based on the configuration file.

        Parameters
        ----------
        config_file : str
            Path to the configuration file.
        update_interval : float, optional
            Time interval in seconds between plot updates (default is 0.1 seconds).

        """

        def wait_for_input():
            input()
            self.enter_pressed = True

        # Initialize secondary thread to wait for user input
        self.enter_pressed = False
        self.input_thread = threading.Thread(target=wait_for_input)
        self.input_thread.daemon = True
        self.input_thread.start()

        # Print instructions message
        print("-" * 80)
        print(" Creating thermodynamic cycle interactive plot")
        print("-" * 80)
        print(f" The current configuration file is: '{configuration_file}'")
        print(" Modify the configuration file and save it to update the plot")
        print(" Try to find a good initial guess for the optimization.")
        print(" Using a feasible initial guess improves optimization convergence.")
        print(" Press 'enter' to continue.")
        print("-" * 80)

        # Update the plot until termination signal
        while not self.enter_pressed:
            # Read the configuration file
            self.load_configuration_file(configuration_file)
            self.fitness(self.x0)
            self.plot_cycle()

            # Wait for the specified interval before updating again
            time.sleep(update_interval)

            # Exit interactive plotting when the user closes the figure
            if not plt.fignum_exists(self.figure.number):
                break

            # Exit interactive plotting when the user presses enter
            if self.enter_pressed:
                break

    def plot_cycle_callback(self, x, iter):
        """
        Plot and save the thermodynamic cycle at different optimization iterations

        This function is intended to be passed as callback argument to the optimization solver
        """
        # Update and plot the cycle diagram
        # self.figure.suptitle(f"Optimization iteration: {iter:03d}", fontsize=14, y=0.90)
        self.plot_cycle()

        # Create a 'results' directory if it doesn't exist
        self.optimization_dir = os.path.join(self.out_dir, "optimization")
        os.makedirs(self.optimization_dir, exist_ok=True)

        # Use the solver's iteration number for the filename
        filename = os.path.join(self.optimization_dir, f"iteration_{iter:03d}.png")
        self.figure.savefig(filename)

    def _get_diagram_default_settings(self, plot_config):
        """
        Merges user-provided plot settings with default settings.

        Parameters
        ----------
        user_settings : dict
            A dictionary of user-defined plot settings.

        Returns
        -------
        dict
            A dictionary containing the merged plot settings.
        """
        default_settings = {
            "x_variable": "s",
            "y_variable": "T",
            "x_scale": "linear",
            "y_scale": "linear",
            "plot_saturation_line": True,
            "plot_critical_point": True,
            "plot_quality_isolines": False,
            "plot_pseudocritical_line": False,
            "plot_triple_point_vap": False,
            "plot_triple_point_liq": False,
            "plot_spinodal_line": False,
            # Add other default settings here
        }

        # Combine the global fluid settings with the "plots" settings
        fluid_settings = self.plot_settings.get("fluid", {})
        fluid_settings = {} if fluid_settings is None else fluid_settings
        plot_config = plot_config | fluid_settings

        # Merge with default values
        return default_settings | plot_config

    def _plot_thermodynamic_diagram(self, ax, plot_config, ax_index=0):
        """
        Plots or updates the thermodynamic diagram on a specified axes.

        This function sets up the axes properties according to the provided plot configuration and
        plots the phase diagram for the specified fluid. It then iterates over all the components in
        the cycle data. For heat exchangers, it plots or updates the processes for both the hot and cold sides.
        For other types of components, it plots or updates the process based on the component's data.

        The function also adjusts the plot limits if it's updating an existing plot,
        ensuring that the axes are scaled correctly to fit the new data.

        Parameters:
        ----------
        ax : matplotlib.axes.Axes
            The axes on which to plot or update the thermodynamic diagram.
        plot_config : dict
            A dictionary containing the plot settings, including variables to be plotted on the x and y axes, and scaling information.
        ax_index : int
            The index of the axes in the figure, used for identifying and updating existing plots.
        """
        # Set up axes properties
        ax.set_xlabel(LABEL_MAPPING[plot_config["x_variable"]])
        ax.set_ylabel(LABEL_MAPPING[plot_config["y_variable"]])
        ax.set_xscale(plot_config["x_scale"])
        ax.set_yscale(plot_config["y_scale"])

        # Plot phase diagram
        self.fluid.plot_phase_diagram(axes=ax, **plot_config)

        # Plot thermodynamic processes
        for name, component in self.cycle_data["components"].items():
            # Handle heat exchanger components in a special way
            if component["type"] == "heat_exchanger":
                for side in ["hot_side", "cold_side"]:
                    self._plot_cycle_process(
                        name + "_" + side, plot_config, ax, ax_index=ax_index
                    )
            else:
                self._plot_cycle_process(name, plot_config, ax, ax_index=ax_index)

        # Adjust plot limits if updating
        ax.relim(visible_only=True)
        ax.autoscale_view()

    def _plot_cycle_process(self, name, plot_settings, ax, ax_index=None):
        """
        Creates or updates the plot elements for a specific cycle process on a given axes.

        This method checks if the plot elements for the specified cycle process already exist on the given axes.
        If they exist and an axis index is provided, it updates these elements with new data. Otherwise, it creates
        new plot elements (lines and points) and stores them for future updates.

        Parameters:
        ----------
        name : str
            The name of the cycle process to plot or update.
        plot_settings : dict
            The plot settings dictionary containing settings such as x and y variables.
        ax : matplotlib.axes.Axes
            The axes on which to plot or update the cycle process.
        ax_index : int, optional
            The index of the axes in the figure, used for updating existing plots. If None, new plots are created.
        """

        # Retrieve component data
        x_data, y_data, color = self._get_process_data(
            name, plot_settings["x_variable"], plot_settings["y_variable"]
        )

        # Initialize the dictionary for this axis index if it does not exist
        if ax_index is not None:
            if ax_index not in self.graphics["process_lines"]:
                self.graphics["process_lines"][ax_index] = {}
            if ax_index not in self.graphics["state_points"]:
                self.graphics["state_points"][ax_index] = {}

        # Handle existing plot elements
        if ax_index is not None and name in self.graphics["process_lines"][ax_index]:
            if x_data is None or y_data is None:
                # Hide existing plot elements if data is None
                self.graphics["process_lines"][ax_index][name].set_visible(False)
                self.graphics["state_points"][ax_index][name].set_visible(False)
            else:
                # Update existing plot elements with new data
                self.graphics["process_lines"][ax_index][name].set_data(x_data, y_data)
                self.graphics["state_points"][ax_index][name].set_data(
                    [x_data[0], x_data[-1]], [y_data[0], y_data[-1]]
                )
                self.graphics["process_lines"][ax_index][name].set_visible(True)
                self.graphics["state_points"][ax_index][name].set_visible(True)
        elif x_data is not None and y_data is not None:
            # Create new plot elements if data is not None
            (line,) = ax.plot(
                x_data,
                y_data,
                linestyle="-",
                linewidth=1.25,
                marker="none",
                markersize=4.0,
                markeredgewidth=1.25,
                markerfacecolor="w",
                color=color,
                label=name,
                zorder=1,
            )
            (points,) = ax.plot(
                [x_data[0], x_data[-1]],
                [y_data[0], y_data[-1]],
                linestyle="none",
                linewidth=1.25,
                marker="o",
                markersize=4.0,
                markeredgewidth=1.25,
                markerfacecolor="w",
                color=color,
                zorder=2,
            )

            # Store the new plot elements
            if ax_index is not None:
                self.graphics["process_lines"][ax_index][name] = line
                self.graphics["state_points"][ax_index][name] = points

    def _get_process_data(self, name, prop_x, prop_y):
        """
        Retrieve thermodynamic data for a specified process of the cycle.

        Parameters:
        ----------
        name : str
            Name of the cycle process.
        prop_x : str
            Property name to plot on the x-axis.
        prop_y : str
            Property name to plot on the y-axis.

        Returns:
        -------
        tuple of (np.ndarray, np.ndarray, str)
            x_data: Array of data points for the x-axis.
            y_data: Array of data points for the y-axis.
            color: Color code for the plot.
        """

        # Get heat exchanger side data
        if "_hot_side" in name or "_cold_side" in name:
            # Extract the component and side from the name
            if "_hot_side" in name:
                component_name, side_1 = name.replace("_hot_side", ""), "hot_side"
                side_2 = "cold_side"
            else:
                component_name, side_1 = name.replace("_cold_side", ""), "cold_side"
                side_2 = "hot_side"

            data = self.cycle_data["components"][component_name][side_1]
            data_other_side = self.cycle_data["components"][component_name][side_2]
            is_heat_exchanger = True

        else:  # Handle non-heat exchanger components
            data = self.cycle_data["components"][name]
            is_heat_exchanger = False

        # Retrieve data
        if data["states"]["identifier"][0] == "working_fluid":
            # Components of the cycle
            x_data = data["states"][prop_x]
            y_data = data["states"][prop_y]
        elif is_heat_exchanger and prop_y == "T" and prop_x in ["h", "s"]:
            # Special case for heat exchangers
            x_data = data_other_side["states"][prop_x]
            y_data = data["states"][prop_y]
        else:
            # Other cases
            x_data = None
            y_data = None

        color = data["color"]

        return x_data, y_data, color

    def _plot_pinch_point_diagram(self, ax, ax_index):
        """
        Plots or updates the pinch point diagram for the thermodynamic cycle's heat exchangers.

        This method visualizes the temperature vs. heat flow rate for each heat exchanger in the cycle.
        The function is capable of plotting the diagram for the first time or updating it with the latest data
        if called subsequently. It uses a sorted approach, beginning with the heat exchanger that has the minimum
        temperature on the cold side and proceeding in ascending order of temperature.

        Parameters:
        ----------
        ax : matplotlib.axes.Axes
            The axes on which to plot or update the pinch point diagram.
        ax_index : int
            The index of the axes in the figure, used to identify and access the specific axes
            for updating the existing plot elements stored in the 'graphics' attribute.

        Notes:
        -----
        The method relies on the 'graphics' attribute of the class to store and update plot elements.
        It handles the creation of new plot elements (lines, endpoints, vertical lines) when first called
        and updates these elements with new data from 'cycle_data' during subsequent calls.
        """
        ax.set_xlabel(LABEL_MAPPING["heat"])
        ax.set_ylabel(LABEL_MAPPING["T"])

        # Initialize the graphic object dict for this axis index if it does not exist
        if ax_index not in self.graphics["pinch_point_lines"]:
            self.graphics["pinch_point_lines"][ax_index] = {}

        # Set the axes labels
        ax.set_xlabel(LABEL_MAPPING["heat"])
        ax.set_ylabel(LABEL_MAPPING["T"])

        # Extract heat exchanger names and their minimum cold side temperatures
        heat_exchangers = [
            (name, min(component["cold_side"]["states"]["T"]))
            for name, component in self.cycle_data["components"].items()
            if component["type"] == "heat_exchanger"
        ]

        # Sort heat exchangers by minimum temperature on the cold side
        sorted_heat_exchangers = sorted(heat_exchangers, key=lambda x: x[1])

        # Loop over all heat exchangers
        Q0 = 0.00
        for HX_name, _ in sorted_heat_exchangers:
            component = self.cycle_data["components"][HX_name]
            # Hot side
            c_hot = component["hot_side"]["color"]
            props_hot = component["hot_side"]["states"]
            mass_flow_hot = component["hot_side"]["mass_flow"]
            Q_hot = (props_hot["h"] - props_hot["h"][0]) * mass_flow_hot
            T_hot = props_hot["T"]

            # Cold side
            c_cold = component["cold_side"]["color"]
            props_cold = component["cold_side"]["states"]
            mass_flow_cold = component["cold_side"]["mass_flow"]
            Q_cold = (props_cold["h"] - props_cold["h"][0]) * mass_flow_cold
            T_cold = props_cold["T"]

            # Check if the plot elements for this component already exist
            if HX_name in self.graphics["pinch_point_lines"][ax_index]:
                # Update existing plot elements
                plot_elements = self.graphics["pinch_point_lines"][ax_index][HX_name]
                plot_elements["hot_line"].set_data(Q0 + Q_hot, T_hot)
                plot_elements["cold_line"].set_data(Q0 + Q_cold, T_cold)

                # Update endpoints
                plot_elements["hot_start"].set_data(Q0 + Q_hot[0], T_hot[0])
                plot_elements["hot_end"].set_data(Q0 + Q_hot[-1], T_hot[-1])
                plot_elements["cold_start"].set_data(Q0 + Q_cold[0], T_cold[0])
                plot_elements["cold_end"].set_data(Q0 + Q_cold[-1], T_cold[-1])

                # Update vertical lines
                plot_elements["start_line"].set_xdata([Q0, Q0])
                plot_elements["end_line"].set_xdata([Q0 + Q_hot[-1], Q0 + Q_hot[-1]])
            else:
                # Create new plot elements
                (hot_line,) = ax.plot(Q0 + Q_hot, T_hot, color=c_hot)
                (cold_line,) = ax.plot(Q0 + Q_cold, T_cold, color=c_cold)

                # Create endpoints
                param = {"marker": "o", "markersize": 4.0, "markerfacecolor": "white"}
                (hot_1,) = ax.plot(Q0 + Q_hot[0], T_hot[0], color=c_hot, **param)
                (hot_2,) = ax.plot(Q0 + Q_hot[-1], T_hot[-1], color=c_hot, **param)
                (cold_1,) = ax.plot(Q0 + Q_cold[0], T_cold[0], color=c_cold, **param)
                (cold_2,) = ax.plot(Q0 + Q_cold[-1], T_cold[-1], color=c_cold, **param)

                # Create vertical lines
                param = {"color": "black", "linestyle": "-", "linewidth": 0.75}
                start_line = ax.axvline(x=Q0, zorder=1, **param)
                end_line = ax.axvline(x=Q0 + Q_hot[-1], zorder=1, **param)

                # Store new plot elements
                self.graphics["pinch_point_lines"][ax_index][HX_name] = {
                    "hot_line": hot_line,
                    "cold_line": cold_line,
                    "hot_start": hot_1,
                    "hot_end": hot_2,
                    "cold_start": cold_1,
                    "cold_end": cold_2,
                    "start_line": start_line,
                    "end_line": end_line,
                }

            # Update abscissa for the next heat exchanger
            Q0 += Q_hot[-1]

        ax.set_xlim(left=0 - Q0 / 50, right=Q0 + Q0 / 50)
